#
#   Copyright 2026 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import humps
from hopsworks_common import util


if TYPE_CHECKING:
    from hsfs.core.chart import Chart


class Dashboard:
    """Metadata object used to provide Dashboard information."""

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        charts: list[Chart] | None = None,
        **kwargs,
    ):
        self._id = id
        self._name = name
        self._charts = charts

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> list[Dashboard]:
        if json_dict is None:
            return None

        json_decamelized = humps.decamelize(json_dict)

        if isinstance(json_decamelized, list):
            return [cls(**item) for item in json_decamelized]
        return cls(**json_decamelized)

    def to_dict(self):
        return {
            "id": self._id,
            "name": self._name,
            "charts": self._charts,
        }

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @property
    def id(self) -> int | None:
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        self._id = id

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def charts(self) -> list[Chart] | None:
        return self._charts

    @charts.setter
    def charts(self, charts: list[Chart]) -> None:
        self._charts = charts

    def delete(self) -> None:
        """Delete the dashboard from the feature store.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core.dashboard_api import DashboardApi

        DashboardApi().delete_dashboard(self.id)

    def update(self) -> None:
        """Update the dashboard in the feature store.

        Updates the dashboard metadata with the current values of this object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core.dashboard_api import DashboardApi

        DashboardApi().update_dashboard(self)



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from dataclasses import dataclass

import humps
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hsfs import util
from hsfs.core.vector_db_client import VectorDbClient


class SimilarityFunctionType:
    """Enumeration class representing different types of similarity functions.

    Parameters:
        L2: Represents L2 norm similarity function.
        COSINE: Represents cosine similarity function.
        DOT_PRODUCT: Represents dot product similarity function.
    """

    L2: str = "l2_norm"
    COSINE: str = "cosine"
    DOT_PRODUCT: str = "dot_product"

    def __init__(self) -> None:
        # Fix for the doc
        raise NotImplementedError("This class should not be instantiated.")


@dataclass
class HsmlModel:
    """Data class storing the metadata of a hsml model."""

    model_registry_id: int
    model_name: str
    model_version: int

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)

        return cls(
            model_registry_id=json_decamelized.get("model_registry_id"),
            model_name=json_decamelized.get("model_name"),
            model_version=json_decamelized.get("model_version"),
        )

    @classmethod
    def from_model(cls, model):
        return cls(
            model_registry_id=model.model_registry_id,
            model_name=model.name,
            model_version=model.version,
        )

    # should get from backend because of authorisation check (unshared project etc)
    def get_model(self):
        try:
            from hsml.model import Model
        except ModuleNotFoundError as err:
            raise FeatureStoreException(
                "Model is attached to embedding feature but hsml library is not installed."
                "Install hsml library before getting the feature group."
            ) from err

        path_params = [
            "project",
            client.get_instance()._project_id,
            "modelregistries",
            self.model_registry_id,
            "models",
            self.model_name + "_" + str(self.model_version),
        ]
        query_params = {"expand": "trainingdatasets"}

        model_json = client.get_instance()._send_request(
            "GET", path_params, query_params
        )
        return Model.from_response_json(model_json)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "modelRegistryId": self.model_registry_id,
            "modelName": self.model_name,
            "modelVersion": self.model_version,
        }


class EmbeddingFeature:
    """Represents an embedding feature.

    Parameters:
        name: The name of the embedding feature.
        dimension: The dimensionality of the embedding feature.
        similarity_function_type:
            The type of similarity function used for the embedding feature.
            Available functions are `L2`, `COSINE`, and `DOT_PRODUCT`.
            Defaults to `SimilarityFunctionType.L2`.
        model: `hsml.model.Model` A Model in hsml.
        feature_group: The feature group object that contains the embedding feature.
        embedding_index: `EmbeddingIndex` The index for managing embedding features.
    """

    def __init__(
        self,
        name: str = None,
        dimension: int = None,
        similarity_function_type: SimilarityFunctionType = SimilarityFunctionType.L2,
        model=None,
        feature_group=None,
        embedding_index=None,
    ):
        self._name = name
        self._dimension = dimension
        self._similarity_function_type = similarity_function_type
        self._model = model
        self._feature_group = feature_group
        self._embedding_index = embedding_index

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        hsml_model_json = json_decamelized.get("model")
        hsml_model = (
            HsmlModel.from_response_json(hsml_model_json).get_model()
            if hsml_model_json
            else None
        )

        return cls(
            name=json_decamelized.get("name"),
            dimension=json_decamelized.get("dimension"),
            similarity_function_type=json_decamelized.get("similarity_function_type"),
            model=hsml_model,
        )

    @property
    def name(self):
        """str: The name of the embedding feature."""
        return self._name

    @property
    def dimenstion(self):
        """The dimensionality of the embedding feature.

        This one is excluded from the docs as the name is misspelled but kept to avoid breaking the API.
        """
        return self._dimension

    @property
    def dimension(self):
        """int: The dimensionality of the embedding feature."""
        return self._dimension

    @property
    def similarity_function_type(self):
        """SimilarityFunctionType: The type of similarity function used for the embedding feature."""
        return self._similarity_function_type

    @property
    def model(self):
        """hsml.model.Model: The Model in hsml."""
        return self._model

    @property
    def feature_group(self):
        """FeatureGroup: The feature group object that contains the embedding feature."""
        return self._feature_group

    @feature_group.setter
    def feature_group(self, feature_group):
        """Set the feature group object that contains the embedding feature.

        Args:
            feature_group (FeatureGroup): The feature group object.
        """
        self._feature_group = feature_group

    @property
    def embedding_index(self):
        """EmbeddingIndex: The index for managing embedding features."""
        return self._embedding_index

    @embedding_index.setter
    def embedding_index(self, embedding_index):
        """Set the index for managing embedding features.

        Args:
            embedding_index (EmbeddingIndex): The embedding index object.
        """
        self._embedding_index = embedding_index

    def json(self):
        """Serialize the EmbeddingFeature object to a JSON string."""
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """Convert the EmbeddingFeature object to a dictionary.

        Returns:
            `dict`: A dictionary representation of the EmbeddingFeature object.
        """
        d = {
            "name": self._name,
            "dimension": self._dimension,
            "similarityFunctionType": self._similarity_function_type,
        }
        if self._model:
            d["model"] = HsmlModel.from_model(self._model)
        return d

    def __repr__(self):
        return self.json()


class EmbeddingIndex:
    """Represents an index for managing embedding features.

    Parameters:
        index_name:
            The name of the embedding index.
            The name of the project index is used if not provided.
        features: A list of `EmbeddingFeature` objects for the features that contain embeddings that should be indexed for similarity search.
        col_prefix:
            The prefix to be added to column names when using project index.
            It is managed by Hopsworks and should not be provided.

    Example:
        ```python
        embedding_index = EmbeddingIndex()
        embedding_index.add_embedding(name="user_vector", dimension=256)
        embeddings = embedding_index.get_embeddings()
        ```
    """

    def __init__(
        self,
        index_name: str | None = None,
        features: list[EmbeddingFeature] | None = None,
        col_prefix: str | None = None,
    ):
        self._index_name = index_name
        if features is None:
            self._features = {}
        else:
            self._features = {feat.name: feat for feat in features}
        self._feature_group = None
        self._col_prefix = col_prefix
        self._vector_db_client = None

    def add_embedding(
        self,
        name: str,
        dimension: int,
        similarity_function_type: SimilarityFunctionType
        | None = SimilarityFunctionType.L2,
        model=None,
    ):
        """Adds a new embedding feature to the index.

        Example:
            ```python
            embedding_index = EmbeddingIndex()
            embedding_index.add_embedding(name="user_vector", dimension=256)

            # Attach a hsml model to the embedding feature
            embedding_index = EmbeddingIndex()
            embedding_index.add_embedding(name="user_vector", dimension=256, model=hsml_model)
            ```

        Parameters:
            name: The name of the embedding feature.
            dimension: The dimensionality of the embedding feature.
            similarity_function_type: The type of similarity function to be used.
            model: `hsml.model.Model | None` The hsml model used to generate the embedding.
        """
        self._features[name] = EmbeddingFeature(
            name, dimension, similarity_function_type, model=model
        )

    def get_embedding(self, name: str) -> EmbeddingFeature:
        """Get `EmbeddingFeature` associated with the feature name.

        Parameters:
            name: The name of the embedding feature.

        Returns:
            The `EmbeddingFeature` associated with the name.
        """
        feat = self._features.get(name)
        if feat:
            feat.feature_group = self._feature_group
            feat.embedding_index = self
        return feat

    def get_embeddings(self) -> list[EmbeddingFeature]:
        """Returns the list of `EmbeddingFeature` objects associated with the index.

        Returns:
            All embedding features in the index.
        """
        for feat in self._features.values():
            feat.feature_group = self._feature_group
            feat.embedding_index = self
        return self._features.values()

    def count(self, options: map = None) -> int:
        """Count the number of records in the feature group.

        Parameters:
            options:
                The options used for the request to the vector database.
                The keys are attribute values of [`OpensearchRequestOption`][hopsworks_common.core.opensearch.OpensearchRequestOption].

        Returns:
            The number of records in the feature group.

        Raises:
            ValueError: If the feature group is not initialized.
            hopsworks.client.exceptions.FeatureStoreException: If an error occurs during the count operation.
        """
        if self._vector_db_client is None:
            self._vector_db_client = VectorDbClient(self._feature_group.select_all())
        return self._vector_db_client.count(self.feature_group, options=options)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(
            index_name=json_decamelized.get("index_name"),
            features=[
                EmbeddingFeature.from_response_json(f)
                for f in json_decamelized.get("features")
            ],
            col_prefix=json_decamelized.get("col_prefix"),
        )

    @property
    def feature_group(self):
        """FeatureGroup: The feature group object that contains the embedding feature."""
        return self._feature_group

    @feature_group.setter
    def feature_group(self, feature_group):
        """Setter for the feature group object."""
        self._feature_group = feature_group

    @property
    def index_name(self):
        """str: The name of the embedding index."""
        return self._index_name

    @property
    def col_prefix(self):
        """str: The prefix to be added to column names."""
        return self._col_prefix

    def json(self):
        """Serialize the EmbeddingIndex object to a JSON string."""
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict:
        """Convert the EmbeddingIndex object to a dictionary.

        Returns:
            A dictionary representation of the EmbeddingIndex object.
        """
        return {
            "indexName": self._index_name,
            "features": [feature.to_dict() for feature in self._features.values()],
            "colPrefix": self._col_prefix,
        }

    def __repr__(self):
        return self.json()



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
import logging
from typing import Literal, get_args
from urllib.parse import quote

from hopsworks_common import client
from hopsworks_common.search_results import (
    FeatureGroupSearchResult,
    FeatureSearchResult,
    FeaturestoreSearchResult,
    FeatureViewSearchResult,
    TrainingDatasetSearchResult,
)
from hopsworks_common.util import Encoder


DOC_TYPE_ARG = Literal[
    "FEATUREGROUP", "FEATUREVIEW", "TRAININGDATASET", "FEATURE", "ALL"
]


class TagSearchFilter:
    """Filter for searching entities by tag.

    * ``name``  – the tag name as defined by the Hopsworks Admin,
    * ``key``   – the specific tag key within that tag,
    * ``value`` – the value that the tag key must match.
    """

    def __init__(self, name: str, key: str, value: str):
        self._name = name
        self._key = key
        self._value = value

    @property
    def name(self):
        """Name of the tag."""
        return self._name

    @property
    def key(self):
        """Key (property) of the tag."""
        return self._key

    @property
    def value(self):
        """Value of the tag key."""
        return self._value

    def to_dict(self):
        """Convert TagSearchFilter to dictionary."""
        return {"name": self._name, "key": self._key, "value": self._value}

    def json(self):
        return json.dumps(self, cls=Encoder)

    @classmethod
    def from_dict(cls, tag_dict: dict[str, str]) -> TagSearchFilter:
        """Create a TagSearchFilter from a dictionary."""
        try:
            name = tag_dict["name"]
            key = tag_dict["key"]
            value = tag_dict["value"]
        except KeyError as e:
            err_msg = f"Missing required field '{e.args[0]}' for TagSearchFilter."
            raise ValueError(err_msg) from e
        if not isinstance(name, str) or not name:
            raise ValueError(
                "Field 'name' for TagSearchFilter must be a non-empty string."
            )
        if not isinstance(key, str) or not key:
            raise ValueError(
                "Field 'key' for TagSearchFilter must be a non-empty string."
            )
        if not isinstance(value, str) or not value:
            raise ValueError(
                "Field 'value' for TagSearchFilter must be a non-empty string."
            )
        return cls(name=name, key=key, value=value)


class KeywordSearchFilter:
    """Represents a filter for searching by keyword values."""

    def __init__(self, value: str):
        self._value = value

    @property
    def value(self):
        """Value of the keyword."""
        return self._value

    def to_dict(self):
        """Convert KeywordSearchFilter to dictionary."""
        return {"value": self._value}

    def json(self):
        return json.dumps(self, cls=Encoder)

    @classmethod
    def from_dict(cls, keyword_dict: dict[str, str]) -> KeywordSearchFilter:
        """Create a KeywordSearchFilter from a dictionary."""
        if "value" not in keyword_dict or keyword_dict.get("value") in (None, ""):
            raise ValueError(
                "KeywordSearchFilter 'value' field is required and cannot be empty."
            )
        return cls(value=keyword_dict["value"])


class SearchApi:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    def feature_store(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> FeaturestoreSearchResult:
        """Search for feature groups, feature views, training datasets and features.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `FeaturestoreSearchResult`: The search results containing lists of metadata objects for feature groups, feature views, training datasets, and features.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        search_api = project.get_search_api()

        # Simple search
        result = search_api.feature_store("search-term")

        # Access results
        for fg_meta in result.feature_groups:
            print(f"Feature Group: {fg_meta.name} v{fg_meta.version}")
            print(f"Description: {fg_meta.description}")
            print(f"Highlights: {fg_meta.highlights}")

            # Get the same FeatureGroup object as returned by featurestore.get_feature_group
            fg = fg_meta.get()

        # Search with a single keyword (string)
        result = search_api.feature_store("search-term", keyword_filter="ml")

        # Search with multiple keywords (array of strings)
        result = search_api.feature_store("search-term", keyword_filter=["ml", "production"])

        # Search with tag filter as a single dictionary
        result = search_api.feature_store(
            "search-term",
            tag_filter={"name": "tag1", "key": "environment", "value": "production"}
        )

        # Search with tag filter as an array of dictionaries
        result = search_api.feature_store(
            "search-term",
            tag_filter=[
                {"name": "tag1", "key": "environment", "value": "production"},
                {"name": "tag2", "key": "version", "value": "v1.0"}
            ]
        )

        # Search with TagSearchFilter objects
        from hopsworks_common.core.search_api import TagSearchFilter
        tags = [
            TagSearchFilter(name="tag1", key="environment", value="production"),
            TagSearchFilter(name="tag2", key="version", value="v1.0")
        ]
        result = search_api.feature_store("search-term", tag_filter=tags)

        # Search with both keyword_filter and tag_filter
        result = search_api.feature_store(
            "search-term",
            keyword_filter=["ml", "production"],
            tag_filter=tags
        )
        ```
        """
        return self._search(
            doc_type="ALL",
            search_term=search_term,
            keyword_filter=keyword_filter,
            tag_filter=tag_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )

    def feature_groups(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[FeatureGroupSearchResult]:
        """Search for feature groups only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for feature groups matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        search_api = project.get_search_api()

        # Search for feature groups
        fg_metas = search_api.feature_groups("customer")

        for fg_meta in fg_metas:
            print(f"Feature Group: {fg_meta.name} v{fg_meta.version}")

            # Get the same FeatureGroup object as returned by featurestore.get_feature_group
            fg = fg_meta.get()
        ```
        """
        result = self._search(
            doc_type="FEATUREGROUP",
            search_term=search_term,
            keyword_filter=keyword_filter,
            tag_filter=tag_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )
        return result.feature_groups

    def feature_views(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[FeatureViewSearchResult]:
        """Search for feature views only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for feature views matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        search_api = project.get_search_api()

        # Search for feature views
        fv_metas = search_api.feature_views("customer")

        for fv_meta in fv_metas:
            print(f"Feature View: {fv_meta.name} v{fv_meta.version}")

            # Get the same FeatureView object as returned by featurestore.get_feature_view
            fv = fv_meta.get()
        ```
        """
        result = self._search(
            doc_type="FEATUREVIEW",
            search_term=search_term,
            keyword_filter=keyword_filter,
            tag_filter=tag_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )
        return result.feature_views

    def training_datasets(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[TrainingDatasetSearchResult]:
        """Search for training datasets only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for training datasets matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        search_api = project.get_search_api()

        # Search for training datasets
        td_metas = search_api.training_datasets("model")

        for td_meta in td_metas:
            print(f"Training Dataset: {td_meta.name} v{td_meta.version}")

            # Get the same TrainingDataset object as returned by featurestore.get_training_dataset
            td = td_meta.get()
        ```
        """
        result = self._search(
            doc_type="TRAININGDATASET",
            search_term=search_term,
            keyword_filter=keyword_filter,
            tag_filter=tag_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )
        return result.training_datasets

    def features(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[FeatureSearchResult]:
        """Search for features only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of features matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        search_api = project.get_search_api()

        # Search for features
        features = search_api.features("age")

        for feature in features:
            print(f"Feature: {feature.name}")
        ```
        """
        result = self._search(
            doc_type="FEATURE",
            search_term=search_term,
            keyword_filter=keyword_filter,
            tag_filter=tag_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )
        return result.features

    def _parse_keyword_filter(
        self, keyword_filter: str | list[str] | None
    ) -> list[KeywordSearchFilter] | None:
        """Parse keyword_filter input to a list of KeywordSearchFilter objects.

        Parameters:
            keyword_filter: The keyword filter input to parse.
                - None: returns None
                - Single string: converts to [KeywordSearchFilter]
                - List of strings: converts each to KeywordSearchFilter

        Returns:
            `list[KeywordSearchFilter] | None`: Parsed list of KeywordSearchFilter objects, or None.
        """
        if keyword_filter is None:
            return None

        # If single string, convert to list
        if isinstance(keyword_filter, str):
            keyword_filter = [keyword_filter]

        # Convert all items to KeywordSearchFilter objects
        parsed_keywords = []
        for item in keyword_filter:
            if not isinstance(item, str):
                raise ValueError(
                    f"Invalid keyword filter item. Expected string, got {type(item)}"
                )
            parsed_keywords.append(KeywordSearchFilter(value=item))
        return parsed_keywords

    def _parse_tag_filter(
        self, tag_filter: dict[str, str] | list[dict[str, str] | TagSearchFilter] | None
    ) -> list[TagSearchFilter] | None:
        """Parse tag_filter input to a list of TagSearchFilter objects.

        Parameters:
            tag_filter: The tag filter input to parse.
                - None: returns None
                - Single dict: converts to [TagSearchFilter]
                - List of dicts: converts to [TagSearchFilter]
                - List of TagSearchFilter: returns as-is
                - List of mixed dicts and TagSearchFilter: parses all to TagSearchFilter

        Returns:
            `list[TagSearchFilter] | None`: Parsed list of TagSearchFilter objects, or None.
        """
        if tag_filter is None:
            return None

        # If single dictionary, convert to list
        if isinstance(tag_filter, dict):
            tag_filter = [tag_filter]

        # Convert all items to TagSearchFilter objects
        parsed_tags = []
        for tag in tag_filter:
            if isinstance(tag, TagSearchFilter):
                parsed_tags.append(tag)
            elif isinstance(tag, dict):
                parsed_tags.append(TagSearchFilter.from_dict(tag))
            else:
                raise ValueError(
                    f"Invalid tag filter item. Expected dict or TagSearchFilter, got {type(tag)}"
                )

        return parsed_tags

    def _search(
        self,
        doc_type: DOC_TYPE_ARG,
        search_term: str | None = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | TagSearchFilter]
        | None = None,
        offset: int | None = None,
        limit: int | None = None,
        global_search: bool = False,
    ) -> FeaturestoreSearchResult:
        if doc_type not in get_args(DOC_TYPE_ARG):
            raise ValueError(
                f"doc_type must be one of the following {get_args(DOC_TYPE_ARG)}."
            )

        # Parse keyword_filter to list of KeywordSearchFilter objects
        parsed_keywords = self._parse_keyword_filter(keyword_filter)

        # Parse tag_filter to list of TagSearchFilter objects
        parsed_tags = self._parse_tag_filter(tag_filter)

        if search_term is None and not parsed_keywords and not parsed_tags:
            raise ValueError(
                "At least one of search_term, keyword_filter, or tag_filter must be provided."
            )

        _client = client.get_instance()
        if global_search:
            path_params = ["elastic", "featurestore"]
        else:
            path_params = ["project", _client._project_id, "elastic", "featurestore"]

        headers = {"content-type": "application/json"}
        query_params = {
            "docType": doc_type,
            "from": offset,
            "size": limit,
        }

        if search_term is not None:
            query_params["searchTerm"] = search_term

        if parsed_keywords:
            # Convert list of KeywordSearchFilter objects to list of dictionaries
            keywords_dict = [keyword.to_dict() for keyword in parsed_keywords]
            # Serialize to JSON string and URL-encode all characters (safe="") since JSON contains special chars
            query_params["keywords"] = quote(json.dumps(keywords_dict), safe="")

        if parsed_tags:
            # Convert list of TagSearchFilter objects to list of dictionaries
            tags_dict = [tag.to_dict() for tag in parsed_tags]
            # Serialize to JSON string and URL-encode all characters (safe="") since JSON contains special chars
            query_params["tags"] = quote(json.dumps(tags_dict), safe="")

        result = _client._send_request(
            "GET", path_params, query_params=query_params, headers=headers
        )

        return FeaturestoreSearchResult(result)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from datetime import date, datetime  # noqa: TC003

from pydantic import BaseModel


class FeatureGroup(BaseModel):
    id: int
    name: str
    version: int
    description: str | None = None
    location: str | None = None
    event_time: str | int | date | datetime | None = None
    online_enabled: bool | None = None
    topic_name: str | None = None
    notification_topic_name: str | None = None
    deprecated: bool | None = None

class Feature(BaseModel):
    name: str
    type: str | None = None
    description: str | None = None
    primary: bool | None = None
    event_time: bool | None = None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import contextlib
import json
from typing import Any

import humps
from hopsworks_common import util


class Tag:
    NOT_FOUND_ERROR_CODE = 370002

    def __init__(
        self,
        name: str,
        value: dict[str, Any] | str,
        schema=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        type=None,
        **kwargs,
    ):
        self._name = name
        self._value = value

    def to_dict(self):
        # Backend expects value to always be a string
        # If value is a dict, serialize it to JSON string
        value = self._value
        if isinstance(value, dict):
            value = json.dumps(value)
        return {
            "name": self._name,
            "value": value,
        }

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @staticmethod
    def normalize(
        tags: Tag | dict[str, Any] | list[Tag | dict[str, Any]] | None,
    ) -> list[Tag]:
        """Normalize tags input to a list of Tag objects.

        # Arguments
            tags: Tags in various formats (single Tag, dict, or list of Tags/dicts), or None

        # Returns
            `List[Tag]`: List of Tag objects
        """
        normalized_tags = []
        if tags is not None:
            if isinstance(tags, list):
                for t in tags:
                    if isinstance(t, Tag):
                        normalized_tags.append(t)
                    elif isinstance(t, dict):
                        normalized_tags.append(Tag(**t))
            elif isinstance(tags, Tag):
                normalized_tags.append(tags)
            elif isinstance(tags, dict):
                normalized_tags.append(Tag(**tags))
        return normalized_tags

    @staticmethod
    def tags_to_dict(tags: list[Tag] | None) -> dict[str, Any] | None:
        """Convert a list of tags to API format with count and items.

        # Arguments
            tags: List of Tag objects, or None

        # Returns
            `dict`: Dictionary with count and items, or None if no tags
        """
        if not tags:
            return None
        return {"count": len(tags), "items": [t.to_dict() for t in tags]}

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" not in json_decamelized or json_decamelized["count"] == 0:
            return []
        tags = []
        for tag_dict in json_decamelized["items"]:
            # Skip tags that don't have name and value (e.g., schema-only entries)
            if "name" not in tag_dict or "value" not in tag_dict:
                continue

            # Try to deserialize value from JSON string if it's a string
            if isinstance(tag_dict["value"], str):
                with contextlib.suppress(json.JSONDecodeError, ValueError):
                    tag_dict["value"] = json.loads(tag_dict["value"])

            # Only pass name and value to avoid issues with extra fields like schema
            tags.append(cls(name=tag_dict["name"], value=tag_dict["value"]))
        return tags

    @property
    def name(self):
        """Name of the tag."""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def value(self):
        """Value of tag."""
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"Tag({self._name!r}, {self._value!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import logging
import os
import time
import uuid
from datetime import datetime
from typing import TYPE_CHECKING

from hopsworks_common.client.exceptions import JobExecutionException, RestAPIError
from hopsworks_common.core import dataset_api, execution_api


if TYPE_CHECKING:
    from hopsworks_common.execution import Execution


class ExecutionEngine:
    def __init__(self):
        self._dataset_api = dataset_api.DatasetApi()
        self._execution_api = execution_api.ExecutionApi()
        self._log = logging.getLogger(__name__)

    def download_logs(
        self, execution: Execution, path: str | None = None
    ) -> tuple[str | None, str | None]:
        """Download execution logs to current directory.

        Parameters:
            execution: execution to download logs for
            path: path to download the logs

        Returns:
            downloaded stdout and stderr log path

        Raises:
            JobExecutionException: if path is provided but does not exist.
        """
        if path is not None and not os.path.exists(path):
            raise JobExecutionException(f"Path {path} does not exist")
        if path is None:
            path = os.getcwd()

        job_logs_dir = f"logs-job-{execution.job_name}-exec-{str(execution.id)}_{str(uuid.uuid4())[:16]}"
        download_log_dir = os.path.join(path, job_logs_dir)

        if not os.path.exists(download_log_dir):
            os.mkdir(download_log_dir)

        out_path = None
        if execution.stdout_path is not None and self._dataset_api.exists(
            execution.stdout_path
        ):
            out_path = self._download_log(execution.stdout_path, download_log_dir)

        err_path = None
        if execution.stderr_path is not None and self._dataset_api.exists(
            execution.stderr_path
        ):
            err_path = self._download_log(execution.stderr_path, download_log_dir)

        return out_path, err_path

    def _download_log(self, path, download_log_dir):
        max_num_retries = 12
        retries = 0
        download_path = None
        while retries < max_num_retries:
            try:
                download_path = self._dataset_api.download(
                    path, download_log_dir, overwrite=True
                )
                break
            except RestAPIError as e:
                if (
                    e.response.json().get("errorCode", "") == 110021
                    and e.response.status_code == 400
                    and retries < max_num_retries
                ):
                    retries += 1
                    time.sleep(5)
                else:
                    raise e
        return download_path

    def wait_until_finished(
        self, job, execution, timeout: float | None = None
    ) -> Execution | None:
        """Wait until execution terminates.

        Parameters:
            job: Job of the execution.
            execution: Execution to monitor.
            timeout:
                The maximum waiting time in seconds, if `None` the waiting time is unbounded.
                **Note**: the actual waiting time may be bigger by approximately 3 seconds.

        Returns:
            The final execution or `None` if the timeout is exceeded.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        start_time = datetime.now()

        def passed():
            return (datetime.now() - start_time).total_seconds()

        MAX_LAG = 3.0

        is_yarn_job = job.job_type is not None and (
            job.job_type.lower() == "spark"
            or job.job_type.lower() == "pyspark"
            or job.job_type.lower() == "flink"
        )

        updated_execution = self._execution_api._get(job, execution.id)
        execution_state = None
        while updated_execution.success is None:
            if timeout and timeout <= passed() + MAX_LAG:
                self._log.info("The waiting timeout was exceeded.")
                return updated_execution
            time.sleep(MAX_LAG)
            updated_execution = self._execution_api._get(job, execution.id)
            if execution_state != updated_execution.state:
                if is_yarn_job:
                    self._log.info(
                        f"Waiting for execution to finish. Current state: {updated_execution.state}. Final status: {updated_execution.final_status}"
                    )
                else:
                    self._log.info(
                        f"Waiting for execution to finish. Current state: {updated_execution.state}"
                    )
            execution_state = updated_execution.state

        # wait for log files to be aggregated, max 6 minutes
        await_time = 6 * 60.0
        log_aggregation_files_exist = self._dataset_api.exists(
            updated_execution.stdout_path
        ) and self._dataset_api.exists(updated_execution.stderr_path)
        log_aggregation_files_exist_already = log_aggregation_files_exist
        self._log.info("Waiting for log aggregation to finish.")
        while not log_aggregation_files_exist and await_time >= 0:
            if timeout and timeout <= passed() + MAX_LAG:
                break
            await_time -= MAX_LAG
            time.sleep(MAX_LAG)
            updated_execution = self._execution_api._get(job, execution.id)

            log_aggregation_files_exist = self._dataset_api.exists(
                updated_execution.stdout_path
            ) and self._dataset_api.exists(updated_execution.stderr_path)

        if not log_aggregation_files_exist_already and (
            not timeout or timeout > passed() + 5
        ):
            time.sleep(5)  # Helps for log aggregation to flush to filesystem

        if is_yarn_job and not updated_execution.success:
            self._log.error(
                f"Execution failed with status: {updated_execution.final_status}. See the logs for more information."
            )
        elif not is_yarn_job and not updated_execution.success:
            self._log.error(
                f"Execution failed with status: {updated_execution.state}. See the logs for more information."
            )
        else:
            self._log.info("Execution finished successfully.")

        return updated_execution



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time

from hopsworks_common import client, command, environment, library
from hopsworks_common.client.exceptions import EnvironmentException, RestAPIError


class EnvironmentEngine:
    def await_library_command(self, environment_name, library_name):
        commands = [command.Command(status="ONGOING")]
        while len(commands) > 0 and not self._is_final_status(commands[0]):
            time.sleep(5)
            library = self._poll_commands_library(environment_name, library_name)
            commands = [] if library is None else library._commands

    def await_environment_command(self, environment_name):
        commands = [command.Command(status="ONGOING")]
        while len(commands) > 0 and not self._is_final_status(commands[0]):
            time.sleep(5)
            environment = self._poll_commands_environment(environment_name)
            commands = [] if environment is None else environment._commands

    def _is_final_status(self, command):
        if command.status == "FAILED":
            raise EnvironmentException(
                f"Command failed with stacktrace: \n{command.error_message}"
            )
        return command.status == "SUCCESS"

    def _poll_commands_library(self, environment_name, library_name):
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "python",
            "environments",
            environment_name,
            "libraries",
            library_name,
        ]

        query_params = {"expand": "commands"}
        headers = {"content-type": "application/json"}

        try:
            return library.Library.from_response_json(
                _client._send_request(
                    "GET", path_params, headers=headers, query_params=query_params
                ),
            )
        except RestAPIError as e:
            if (
                e.response.json().get("errorCode", "") == 300003
                and e.response.status_code == 404
            ):
                return None

    def _poll_commands_environment(self, environment_name):
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "python",
            "environments",
            environment_name,
        ]

        query_params = {"expand": "commands"}
        headers = {"content-type": "application/json"}

        return environment.Environment.from_response_json(
            _client._send_request(
                "GET", path_params, headers=headers, query_params=query_params
            ),
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, TypeVar

import humps
from hsfs import (
    expectation_suite,
    feature,
    feature_group,
    feature_view,
    storage_connector,
    tag,
    training_dataset,
    usage,
    util,
)
from hsfs.core import (
    data_source as ds,
)
from hsfs.core import (
    feature_group_api,
    feature_group_engine,
    feature_view_engine,
    search_api,
    storage_connector_api,
    training_dataset_api,
    transformation_function_engine,
)
from hsfs.core.chart import Chart
from hsfs.core.chart_api import ChartApi
from hsfs.core.dashboard import Dashboard
from hsfs.core.dashboard_api import DashboardApi
from hsfs.core.job import Job
from hsfs.decorators import typechecked
from hsfs.transformation_function import TransformationFunction


if TYPE_CHECKING:
    from hopsworks_common.core.constants import HAS_NUMPY, HAS_POLARS

    if HAS_NUMPY:
        import numpy as np
    if HAS_POLARS:
        import polars as pl

    import datetime
    from datetime import timedelta

    import pandas as pd
    from hsfs.constructor.query import Query
    from hsfs.embedding import EmbeddingIndex
    from hsfs.hopsworks_udf import HopsworksUdf
    from hsfs.online_config import OnlineConfig
    from hsfs.statistics_config import StatisticsConfig


@typechecked
class FeatureStore:
    """Feature Store class used to manage feature store entities, like feature groups and feature views."""

    DEFAULT_VERSION = 1

    def __init__(
        self,
        featurestore_id: int,
        featurestore_name: str,
        created: str | datetime.datetime,
        project_name: str,
        project_id: int,
        offline_featurestore_name: str,
        online_enabled: bool,
        num_feature_groups: int | None = None,
        num_training_datasets: int | None = None,
        num_storage_connectors: int | None = None,
        num_feature_views: int | None = None,
        online_featurestore_name: str | None = None,
        online_featurestore_size: int | None = None,
        **kwargs,
    ) -> None:
        self._id = featurestore_id
        self._name = featurestore_name
        self._created = created
        self._project_name = project_name
        self._project_id = project_id
        self._online_feature_store_name = online_featurestore_name
        self._online_feature_store_size = online_featurestore_size
        self._offline_feature_store_name = offline_featurestore_name
        self._online_enabled = online_enabled
        self._num_feature_groups = num_feature_groups
        self._num_training_datasets = num_training_datasets
        self._num_storage_connectors = num_storage_connectors
        self._num_feature_views = num_feature_views

        self._feature_group_api: feature_group_api.FeatureGroupApi = (
            feature_group_api.FeatureGroupApi()
        )
        self._storage_connector_api: storage_connector_api.StorageConnectorApi = (
            storage_connector_api.StorageConnectorApi()
        )
        self._training_dataset_api: training_dataset_api.TrainingDatasetApi = (
            training_dataset_api.TrainingDatasetApi(self._id)
        )

        self._feature_group_engine: feature_group_engine.FeatureGroupEngine = (
            feature_group_engine.FeatureGroupEngine(self._id)
        )

        self._transformation_function_engine: transformation_function_engine.TransformationFunctionEngine = transformation_function_engine.TransformationFunctionEngine(
            self._id
        )
        self._feature_view_engine: feature_view_engine.FeatureViewEngine = (
            feature_view_engine.FeatureViewEngine(self._id)
        )
        self._search_api: search_api.SearchApi = search_api.SearchApi()

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> FeatureStore:
        json_decamelized = humps.decamelize(json_dict)
        # fields below are removed from 3.4. remove them for backward compatibility.
        json_decamelized.pop("hdfs_store_path", None)
        json_decamelized.pop("featurestore_description", None)
        json_decamelized.pop("inode_id", None)
        return cls(**json_decamelized)

    def get_feature_group(
        self, name: str, version: int = None
    ) -> (
        feature_group.FeatureGroup
        | feature_group.ExternalFeatureGroup
        | feature_group.SpineGroup
    ):
        """Get a feature group entity from the feature store.

        Getting a feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            fg = fs.get_feature_group(
                    name="electricity_prices",
                    version=1,
                )
            ```

        Parameters:
            name: Name of the feature group to get.
            version: Version of the feature group to retrieve, defaults to `None` and will return the `version=1`.

        Returns:
            The feature group metadata object or `None` if it does not exist.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        if version is None:
            warnings.warn(
                f"No version provided for getting feature group `{name}`, defaulting to `{self.DEFAULT_VERSION}`.",
                util.VersionWarning,
                stacklevel=1,
            )
            version = self.DEFAULT_VERSION
        feature_group_object = self._feature_group_api.get(self.id, name, version)
        if feature_group_object:
            feature_group_object.feature_store = self
            util.check_missing_mandatory_tags(
                feature_group_object.missing_mandatory_tags
            )
        return feature_group_object

    def get_feature_groups(
        self, name: str | None = None
    ) -> list[
        feature_group.FeatureGroup
        | feature_group.ExternalFeatureGroup
        | feature_group.SpineGroup
    ]:
        """Get all feature groups from the feature store, or all versions of a feature group specified by its name.

        Getting a feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # retrieve all versions of electricity_prices feature group
            fgs_list = fs.get_feature_groups(
                    name="electricity_prices"
                )
            ```

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # retrieve all feature groups available in the feature store
            fgs_list = fs.get_feature_groups()
            ```

        Parameters:
            name: Name of the feature group to get the versions of; by default it is `None` and all feature groups are returned.

        Returns:
            List of feature group metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        if name:
            feature_group_object = self._feature_group_api.get(self.id, name, None)
        else:
            feature_group_object = self._feature_group_api.get_all(self.id)
        for fg_object in feature_group_object:
            fg_object.feature_store = self
        return feature_group_object

    @usage.method_logger
    def get_on_demand_feature_group(
        self, name: str, version: int = None
    ) -> feature_group.ExternalFeatureGroup:
        """Get an external feature group entity from the feature store.

        Warning: Deprecated
            `get_on_demand_feature_group` method is deprecated.
            Use the `get_external_feature_group` method instead.

        Getting an external feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Parameters:
            name: Name of the external feature group to get.
            version: Version of the external feature group to retrieve, defaults to `None` and will return the `version=1`.

        Returns:
            The external feature group metadata object or `None` if it does not exist.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self.get_external_feature_group(name, version)

    @usage.method_logger
    def get_external_feature_group(
        self, name: str, version: int = None
    ) -> feature_group.ExternalFeatureGroup:
        """Get an external feature group entity from the feature store.

        Getting an external feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            external_fg = fs.get_external_feature_group("external_fg_test")
            ```

        Parameters:
            name: Name of the external feature group to get.
            version: Version of the external feature group to retrieve, by defaults to `None` and will return the `version=1`.

        Returns:
            The external feature group metadata object or `None` if it does not exist.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        if version is None:
            warnings.warn(
                f"No version provided for getting feature group `{name}`, defaulting to `{self.DEFAULT_VERSION}`.",
                util.VersionWarning,
                stacklevel=1,
            )
            version = self.DEFAULT_VERSION
        feature_group_object = self._feature_group_api.get(
            self.id,
            name,
            version,
        )
        if feature_group_object:
            feature_group_object.feature_store = self
            util.check_missing_mandatory_tags(
                feature_group_object.missing_mandatory_tags
            )
        return feature_group_object

    @usage.method_logger
    def get_on_demand_feature_groups(
        self, name: str
    ) -> list[feature_group.ExternalFeatureGroup]:
        """Get a list of all versions of an external feature group entity from the feature store.

        Warning: Deprecated
            `get_on_demand_feature_groups` method is deprecated.
            Use the `get_external_feature_groups` method instead.

        Getting an external feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Parameters:
            name: Name of the external feature group to get.

        Returns:
            List of external feature group metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self.get_external_feature_groups(name)

    @usage.method_logger
    def get_external_feature_groups(
        self, name: str | None = None
    ) -> list[feature_group.ExternalFeatureGroup]:
        """Get a list of all external feature groups from the feature store, or all versions of an external feature group.

        Getting an external feature group from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame or use the `Query`-API to perform joins between feature groups.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            external_fgs_list = fs.get_external_feature_groups("external_fg_test")
            ```

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # retrieve all external feature groups available in the feature store
            external_fgs_list = fs.get_external_feature_groups()
            ```

        Parameters:
            name: Name of the external feature group to get the versions of; by default it is `None` and all external feature groups are returned.

        Returns:
            List of external feature group metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        fgs = self.get_feature_groups(name)
        return [fg for fg in fgs if isinstance(fg, feature_group.ExternalFeatureGroup)]

    def get_training_dataset(
        self, name: str, version: int = None
    ) -> training_dataset.TrainingDataset:
        """Get a training dataset entity from the feature store.

        Warning: Deprecated
            `TrainingDataset` is deprecated, use `FeatureView` instead.
            You can still retrieve old training datasets using this method, but after upgrading the old training datasets will also be available under a Feature View with the same name and version.

            It is recommended to use this method only for old training datasets that have been created directly from Dataframes and not with Query objects.

        Getting a training dataset from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame.

        Parameters:
            name: Name of the training dataset to get.
            version: Version of the training dataset to retrieve, defaults to `None` and will return the `version=1`.

        Returns:
            The training dataset metadata object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        if version is None:
            warnings.warn(
                f"No version provided for getting training dataset `{name}`, defaulting to `{self.DEFAULT_VERSION}`.",
                util.VersionWarning,
                stacklevel=1,
            )
            version = self.DEFAULT_VERSION
        training_dataset_object = self._training_dataset_api.get(name, version)
        if training_dataset_object:
            util.check_missing_mandatory_tags(
                training_dataset_object.missing_mandatory_tags
            )
        return training_dataset_object

    def get_training_datasets(
        self, name: str
    ) -> list[training_dataset.TrainingDataset]:
        """Get a list of all versions of a training dataset entity from the feature store.

        Warning: Deprecated
            `TrainingDataset` is deprecated, use `FeatureView` instead.

        Getting a training dataset from the Feature Store means getting its metadata handle so you can subsequently read the data into a Spark or Pandas DataFrame.

        Parameters:
            name: Name of the training dataset to get.

        Returns:
            List of training dataset metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._training_dataset_api.get(name, None)

    @usage.method_logger
    def get_storage_connector(self, name: str) -> storage_connector.StorageConnector:
        """Get a previously created storage connector from the feature store.

        Storage connectors encapsulate all information needed for the execution engine to read and write to specific storage.
        This storage can be S3, a JDBC compliant database or the distributed filesystem HOPSFS.

        If you want to connect to the online feature store, see the `get_online_storage_connector` method to get the JDBC connector for the Online Feature Store.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            sc = fs.get_storage_connector("demo_fs_meb10000_Training_Datasets")
            ```

        Parameters:
            name: Name of the storage connector to retrieve.

        Returns:
            Storage connector object.
        """
        return self._storage_connector_api.get(self._id, name)

    def sql(
        self,
        query: str,
        dataframe_type: Literal[
            "default", "spark", "pandas", "polars", "numpy", "python"
        ] = "default",
        online: bool = False,
        read_options: dict | None = None,
    ) -> pd.DataFrame | pd.Series | np.ndarray | pl.DataFrame:
        """Execute SQL command on the offline or online feature store database.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # construct the query and show head rows
            query_res_head = fs.sql("SELECT * FROM `fg_1`").head()
            ```

        Parameters:
            query: The SQL query to execute.
            dataframe_type:
                The type of the returned dataframe.
                Defaults to `"default"`, which maps to Spark dataframe for the Spark Engine and Pandas dataframe for the Python engine.
            online: Set to true to execute the query against the online feature store.
            read_options:
                Additional options as key/value pairs to pass to the execution engine.

                For spark engine: Dictionary of read options for Spark.

                For python engine:
                If running queries on the online feature store, users can provide an entry `{'external': True}`, this instructs the library to use the `host` parameter in the [`hopsworks.login`][hopsworks.login] to establish the connection to the online feature store.
                If not set, or set to False, the online feature store storage connector is used which relies on the private ip.

        Returns:
            DataFrame depending on the chosen type.
        """
        return self._feature_group_engine.sql(
            query, self._name, dataframe_type, online, read_options or {}
        )

    @usage.method_logger
    def get_online_storage_connector(self) -> storage_connector.StorageConnector:
        """Get the storage connector for the Online Feature Store of the respective project's feature store.

        The returned storage connector depends on the project that you are connected to.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            online_storage_connector = fs.get_online_storage_connector()
            ```

        Returns:
            JDBC storage connector to the Online Feature Store.
        """
        return self._storage_connector_api.get_online_connector(self._id)

    def _normalize_tags(
        self,
        tags: tag.Tag | dict[str, Any] | list[tag.Tag | dict[str, Any]] | None,
    ) -> list[tag.Tag]:
        """Normalize tags input to a list of Tag objects.

        # Arguments
            tags: Tags in various formats (single Tag, dict, or list of Tags/dicts)

        # Returns
            `list[tag.Tag]`: List of Tag objects.
        """
        return tag.Tag.normalize(tags)

    @usage.method_logger
    def create_feature_group(
        self,
        name: str,
        version: int | None = None,
        description: str = "",
        online_enabled: bool = False,
        time_travel_format: str | None = None,
        partition_key: list[str] | None = None,
        primary_key: list[str] | None = None,
        foreign_key: list[str] | None = None,
        embedding_index: EmbeddingIndex | None = None,
        hudi_precombine_key: str | None = None,
        features: list[feature.Feature] | None = None,
        statistics_config: StatisticsConfig | bool | dict | None = None,
        event_time: str | None = None,
        stream: bool = False,
        expectation_suite: expectation_suite.ExpectationSuite
        | TypeVar("great_expectations.core.ExpectationSuite")
        | None = None,
        parents: list[feature_group.FeatureGroup] | None = None,
        topic_name: str | None = None,
        notification_topic_name: str | None = None,
        transformation_functions: list[TransformationFunction | HopsworksUdf]
        | None = None,
        online_config: OnlineConfig | dict[str, Any] | None = None,
        offline_backfill_every_hr: int | str | None = None,
        storage_connector: storage_connector.StorageConnector | dict[str, Any] = None,
        path: str | None = None,
        data_source: ds.DataSource | dict[str, Any] | None = None,
        ttl: float | timedelta | None = None,
        ttl_enabled: bool | None = None,
        online_disk: bool | None = None,
        tags: tag.Tag | dict[str, Any] | list[tag.Tag | dict[str, Any]] | None = None,
    ) -> feature_group.FeatureGroup:
        """Create a feature group metadata object.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # define the on-demand transformation functions
            @udf(int)
            def plus_one(value):
                return value + 1

            @udf(int)
            def plus_two(value):
                return value + 2

            # construct list of "transformation functions" on features
            transformation_functions = [plus_one("feature1"), plus_two("feature2")]

            fg = fs.create_feature_group(
                name='air_quality',
                description='Air Quality characteristics of each day',
                version=1,
                primary_key=['city','date'],
                online_enabled=True,
                event_time='date',
                transformation_functions=transformation_functions,
                online_config={'online_comments': ['NDB_TABLE=READ_BACKUP=1']},
                online_disk=True,  # Online data will be stored on disk instead of in memory
                ttl=timedelta(days=7)  # features will be deleted after 7 days
            )
            ```

        Note: Lazy
            This method is lazy and does not persist any metadata or feature data in the feature store on its own.
            To persist the feature group and save feature data along the metadata in the feature store, call the `save()` method with a DataFrame.

        Parameters:
            name: Name of the feature group to create.
            version: Version of the feature group to create, defaults to `None` and will create the feature group with incremented version from the last version in the feature store.
            description: A string describing the contents of the feature group to improve discoverability for Data Scientists.
            online_enabled: Define whether the feature group should be made available also in the online feature store for low latency access.
            time_travel_format: Format used for time travel, defaults to `"HUDI"`.
            partition_key: A list of feature names to be used as partition key when writing the feature data to the offline storage, defaults to empty list `[]`.
            primary_key:
                A list of feature names to be used as primary key for the feature group.
                This primary key can be a composite key of multiple features and will be used as joining key, if not specified otherwise.
                Defaults to empty list `[]`, and the feature group won't have any primary key.
            foreign_key:
                A list of feature names to be used as foreign key for the feature group.
                Foreign key is referencing the primary key of another feature group and can be used as joining key.
                Defaults to empty list `[]`, and the feature group won't have any foreign key.
            embedding_index:
                [`EmbeddingIndex`][hsfs.embedding.EmbeddingIndex]. If an embedding index is provided, vector database is used as online feature store.
                This enables similarity search by using [`FeatureGroup.find_neighbors`][hsfs.feature_group.FeatureGroup.find_neighbors].
            hudi_precombine_key:
                A feature name to be used as a precombine key for the `"HUDI"` feature group.
                If feature group has time travel format `"HUDI"` and hudi precombine key was not specified then the first primary key of the feature group will be used as hudi precombine key.
            features:
                Optionally, define the schema of the feature group manually as a list of `Feature` objects.
                Defaults to empty list `[]` and will use the schema information of the DataFrame provided in the `save` method.
            statistics_config:
                A configuration object, or a dictionary with keys:

                - `enabled` to generally enable descriptive statistics computation for this feature group,
                - `correlations` to turn on feature correlation computation,
                - `histograms` to compute feature value frequencies, and
                - `exact_uniqueness` to compute uniqueness, distinctness and entropy.

                The values should be booleans indicating the setting.
                To fully turn off statistics computation pass `statistics_config=False`.
                By default, it computes only descriptive statistics.
            event_time:
                Optionally, provide the name of the feature containing the event time for the features in this feature group.
                If event_time is set the feature group can be used for point-in-time joins.

                Note: Event time data type restriction
                    The supported data types for the event time column are: `timestamp`, `date` and `bigint`.

            stream:
                Optionally, define whether the feature group should support real time stream writing capabilities.
                Stream enabled Feature Groups have unified single API for writing streaming features transparently to both online and offline store.
            expectation_suite:
                Optionally, attach an expectation suite to the feature group which dataframes should be validated against upon insertion.
            parents:
                Optionally, define the parents of this feature group as the origin where the data is coming from.
            topic_name:
                Optionally, define the name of the topic used for data ingestion.
                If left undefined it defaults to using project topic.
            notification_topic_name:
                Optionally, define the name of the topic used for sending notifications when entries are inserted or updated on the online feature store.
                If left undefined no notifications are sent.
            transformation_functions:
                On-Demand Transformation functions attached to the feature group.
                It can be a list of list of user defined functions defined using the hopsworks `@udf` decorator.
                Defaults to `None`, no transformations.
            online_config: Optionally, define configuration which is used to configure online table.
            offline_backfill_every_hr:
                If specified, the materialization job will be scheduled to run periodically.
                The value can be either an integer representing the number of hours between each run or a string representing a cron expression.
                Set the value to None to avoid scheduling the materialization job.
                By default, no scheduling is done.
            storage_connector: The storage connector used to establish connectivity with the data source.
            path: The location within the scope of the storage connector, from where to read the data for the external feature group.
            data_source:
                The data source specifying the location of the data.
                Overrides the path and query arguments when specified.
            ttl:
                Optional time-to-live duration for features in this group.
                Can be specified as:

                - An integer or float representing seconds
                - A timedelta object

                This ttl value is added to the event time of the feature group and when the system time exceeds the event time + ttl, the entries will be automatically removed.
                The system time zone is in UTC.

                By default, no TTL is set.
            ttl_enabled:
                Optionally, enable TTL for this feature group.
                Defaults to True if ttl is set.
            online_disk:
                Optionally, specify online data storage for this feature group.
                When set to True data will be stored on disk, instead of in memory.
                Overrides online_config.table_space.
                Defaults to using cluster wide configuration 'featurestore_online_tablespace' to identify tablespace for disk storage.
            tags:
                Optionally, define tags for the feature group. Tags can be provided as:
                - A single Tag object
                - A dictionary with 'name' and 'value' keys (e.g., {"name": "tag1", "value": "value1"})
                - A list of Tag objects
                - A list of dictionaries with 'name' and 'value' keys
                Tags will be attached to the feature group after it is saved. Defaults to None.

        Returns:
            The feature group metadata object.
        """
        normalized_tags = self._normalize_tags(tags)

        if not data_source:
            data_source = ds.DataSource(path=path)
        feature_group_object = feature_group.FeatureGroup(
            name=name,
            version=version,
            description=description,
            online_enabled=online_enabled,
            time_travel_format=time_travel_format,
            partition_key=partition_key or [],
            primary_key=primary_key or [],
            foreign_key=foreign_key or [],
            hudi_precombine_key=hudi_precombine_key,
            featurestore_id=self._id,
            featurestore_name=self._name,
            features=features or [],
            embedding_index=embedding_index,
            statistics_config=statistics_config,
            event_time=event_time,
            stream=stream,
            expectation_suite=expectation_suite,
            parents=parents or [],
            topic_name=topic_name,
            notification_topic_name=notification_topic_name,
            transformation_functions=transformation_functions,
            online_config=online_config,
            offline_backfill_every_hr=offline_backfill_every_hr,
            storage_connector=storage_connector,
            data_source=data_source,
            ttl=ttl,
            ttl_enabled=ttl_enabled,
            online_disk=online_disk,
            tags=normalized_tags,
        )
        feature_group_object.feature_store = self
        return feature_group_object

    @usage.method_logger
    def get_or_create_feature_group(
        self,
        name: str,
        version: int,
        description: str | None = "",
        online_enabled: bool | None = False,
        time_travel_format: str | None = None,
        partition_key: list[str] | None = None,
        primary_key: list[str] | None = None,
        foreign_key: list[str] | None = None,
        embedding_index: EmbeddingIndex | None = None,
        hudi_precombine_key: str | None = None,
        features: list[feature.Feature] | None = None,
        statistics_config: StatisticsConfig | bool | dict | None = None,
        expectation_suite: expectation_suite.ExpectationSuite
        | TypeVar("great_expectations.core.ExpectationSuite")
        | None = None,
        event_time: str | None = None,
        stream: bool | None = False,
        parents: list[feature_group.FeatureGroup] | None = None,
        topic_name: str | None = None,
        notification_topic_name: str | None = None,
        transformation_functions: list[TransformationFunction | HopsworksUdf]
        | None = None,
        online_config: OnlineConfig | dict[str, Any] | None = None,
        offline_backfill_every_hr: int | str | None = None,
        storage_connector: storage_connector.StorageConnector | dict[str, Any] = None,
        path: str | None = None,
        data_source: ds.DataSource | dict[str, Any] | None = None,
        ttl: float | timedelta | None = None,
        ttl_enabled: bool | None = None,
        online_disk: bool | None = None,
    ) -> (
        feature_group.FeatureGroup
        | feature_group.ExternalFeatureGroup
        | feature_group.SpineGroup
    ):
        """Get feature group metadata object or create a new one if it doesn't exist.

        This method doesn't update existing feature group metadata object.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            fg = fs.get_or_create_feature_group(
                name="electricity_prices",
                version=1,
                description="Electricity prices from NORD POOL",
                primary_key=["day", "area"],
                online_enabled=True,
                event_time="timestamp",
                transformation_functions=transformation_functions,
                online_config={'online_comments': ['NDB_TABLE=READ_BACKUP=1']},
                online_disk=True, # Online data will be stored on disk instead of in memory
                ttl=timedelta(days=30),
            )
            ```

        Note: Lazy
            This method is lazy and does not persist any metadata or feature data in the feature store on its own.
            To persist the feature group and save feature data along the metadata in the feature store, call the `insert()` method with a DataFrame.

        Parameters:
            name: Name of the feature group to create.
            version: Version of the feature group to retrieve or create.
            description: A string describing the contents of the feature group to improve discoverability for Data Scientists.
            online_enabled: Define whether the feature group should be made available also in the online feature store for low latency access.
            time_travel_format: Format used for time travel, defaults to `"HUDI"`.
            partition_key: A list of feature names to be used as partition key when writing the feature data to the offline storage, defaults to empty list `[]`.
            primary_key:
                A list of feature names to be used as primary key for the feature group.
                This primary key can be a composite key of multiple features and will be used as joining key, if not specified otherwise.
                Defaults to empty list `[]`, and the feature group won't have any primary key.
            foreign_key:
                A list of feature names to be used as foreign key for the feature group.
                Foreign key is referencing the primary key of another feature group and can be used as joining key.
                Defaults to empty list `[]`, and the feature group won't have any foreign key.
            embedding_index:
                [`EmbeddingIndex`][hsfs.embedding.EmbeddingIndex]. If an embedding index is provided, vector database is used as online feature store.
                This enables similarity search by using [`FeatureGroup.find_neighbors`][hsfs.feature_group.FeatureGroup.find_neighbors].
            hudi_precombine_key:
                A feature name to be used as a precombine key for the `"HUDI"` feature group.
                If feature group has time travel format `"HUDI"` and hudi precombine key was not specified then the first primary key of the feature group will be used as hudi precombine key.
            features:
                Optionally, define the schema of the feature group manually as a list of `Feature` objects.
                Defaults to empty list `[]` and will use the schema information of the DataFrame provided in the `save` method.
            statistics_config:
                A configuration object, or a dictionary with keys:

                - `enabled` to generally enable descriptive statistics computation for this feature group,
                - `correlations` to turn on feature correlation computation,
                - `histograms` to compute feature value frequencies, and
                - `exact_uniqueness` to compute uniqueness, distinctness and entropy.

                The values should be booleans indicating the setting.
                To fully turn off statistics computation pass `statistics_config=False`.
                By default, it computes only descriptive statistics.
            event_time:
                Optionally, provide the name of the feature containing the event time for the features in this feature group.
                If event_time is set the feature group can be used for point-in-time joins.

                Note: Event time data type restriction
                    The supported data types for the event time column are: `timestamp`, `date` and `bigint`.

            stream:
                Optionally, define whether the feature group should support real time stream writing capabilities.
                Stream enabled Feature Groups have unified single API for writing streaming features transparently to both online and offline store.
            expectation_suite:
                Optionally, attach an expectation suite to the feature group which dataframes should be validated against upon insertion.
            parents:
                Optionally, define the parents of this feature group as the origin where the data is coming from.
            topic_name:
                Optionally, define the name of the topic used for data ingestion.
                If left undefined it defaults to using project topic.
            notification_topic_name:
                Optionally, define the name of the topic used for sending notifications when entries are inserted or updated on the online feature store.
                If left undefined no notifications are sent.
            transformation_functions:
                On-Demand Transformation functions attached to the feature group.
                It can be a list of list of user defined functions defined using the hopsworks `@udf` decorator.
                Defaults to `None`, no transformations.
            online_config: Optionally, define configuration which is used to configure online table.
            offline_backfill_every_hr:
                If specified, the materialization job will be scheduled to run periodically.
                The value can be either an integer representing the number of hours between each run or a string representing a cron expression.
                Set the value to None to avoid scheduling the materialization job.
                By default, no scheduling is done.
            storage_connector: The storage connector used to establish connectivity with the data source.
            path: The location within the scope of the storage connector, from where to read the data for the external feature group.
            data_source:
                The data source specifying the location of the data.
                Overrides the path and query arguments when specified.
            ttl:
                Optional time-to-live duration for features in this group.
                Can be specified as:

                - An integer or float representing seconds
                - A timedelta object

                This ttl value is added to the event time of the feature group and when the system time exceeds the event time + ttl, the entries will be automatically removed.
                The system time zone is in UTC.

                By default, no TTL is set.
            ttl_enabled:
                Optionally, enable TTL for this feature group.
                Defaults to True if ttl is set.
            online_disk:
                Optionally, specify online data storage for this feature group.
                When set to True data will be stored on disk, instead of in memory.
                Overrides online_config.table_space.
                Defaults to using cluster wide configuration 'featurestore_online_tablespace' to identify tablespace for disk storage.

        Returns:
            The feature group metadata object.
        """
        feature_group_object = self._feature_group_api.get(self.id, name, version)
        if not feature_group_object:
            if not data_source:
                data_source = ds.DataSource(path=path)
            feature_group_object = feature_group.FeatureGroup(
                name=name,
                version=version,
                description=description,
                online_enabled=online_enabled,
                time_travel_format=time_travel_format,
                partition_key=partition_key or [],
                primary_key=primary_key or [],
                foreign_key=foreign_key or [],
                embedding_index=embedding_index,
                hudi_precombine_key=hudi_precombine_key,
                featurestore_id=self._id,
                featurestore_name=self._name,
                features=features or [],
                statistics_config=statistics_config,
                event_time=event_time,
                stream=stream,
                expectation_suite=expectation_suite,
                parents=parents or [],
                topic_name=topic_name,
                notification_topic_name=notification_topic_name,
                transformation_functions=transformation_functions,
                online_config=online_config,
                offline_backfill_every_hr=offline_backfill_every_hr,
                storage_connector=storage_connector,
                data_source=data_source,
                ttl=ttl,
                ttl_enabled=ttl_enabled,
                online_disk=online_disk,
            )
        feature_group_object.feature_store = self
        return feature_group_object

    @usage.method_logger
    def create_on_demand_feature_group(
        self,
        name: str,
        storage_connector: storage_connector.StorageConnector,
        query: str | None = None,
        data_format: str | None = None,
        path: str | None = "",
        options: dict[str, str] | None = None,
        version: int | None = None,
        description: str | None = "",
        primary_key: list[str] | None = None,
        foreign_key: list[str] | None = None,
        features: list[feature.Feature] | None = None,
        statistics_config: StatisticsConfig | bool | dict | None = None,
        event_time: str | None = None,
        expectation_suite: expectation_suite.ExpectationSuite
        | TypeVar("great_expectations.core.ExpectationSuite")
        | None = None,
        topic_name: str | None = None,
        notification_topic_name: str | None = None,
        data_source: ds.DataSource | dict[str, Any] | None = None,
        online_enabled: bool = False,
        ttl: float | timedelta | None = None,
        ttl_enabled: bool | None = None,
    ) -> feature_group.ExternalFeatureGroup:
        """Create an external feature group metadata object.

        Warning: Deprecated
            `create_on_demand_feature_group` method is deprecated.
            Use the `create_external_feature_group` method instead.

        Note: Lazy
            This method is lazy and does not persist any metadata in the
            feature store on its own.
            To persist the feature group metadata in the feature store, call the `save()` method.

        Parameters:
            name: Name of the external feature group to create.
            storage_connector: The storage connector used to establish connectivity with the data source.
            query:
                A string containing a SQL query valid for the target data source.
                The query will be used to pull data from the data sources when the feature group is used.
            data_format: If the external feature groups refers to a directory with data, the data format to use when reading it.
            path: The location within the scope of the storage connector, from where to read the data for the external feature group.
            options:
                Additional options to be used by the engine when reading data from the specified storage connector.
                For example, `{"header": True}` when reading CSV files with column names in the first row.
            version: Version of the external feature group to retrieve, defaults to `None` and will create the feature group with incremented version from the last version in the feature store.
            description: A string describing the contents of the external feature group to improve discoverability for Data Scientists.
            primary_key:
                A list of feature names to be used as primary key for the feature group.
                This primary key can be a composite key of multiple features and will be used as joining key, if not specified otherwise.
                Defaults to empty list `[]`, and the feature group won't have any primary key.
            foreign_key:
                A list of feature names to be used as foreign key for the feature group.
                Foreign key is referencing the primary key of another feature group and can be used as joining key.
                Defaults to empty list `[]`, and the feature group won't have any foreign key.
            features:
                Optionally, define the schema of the external feature group manually as a list of `Feature` objects.
                Defaults to empty list `[]` and will use the schema information of the DataFrame resulting by executing the provided query against the data source.
            statistics_config:
                A configuration object, or a dictionary with keys:

                - `"enabled"` to generally enable descriptive statistics computation for this external feature group,
                - `"correlations"` to turn on feature correlation
                computation,
                - `"histograms"` to compute feature value frequencies, and
                - `"exact_uniqueness"` to compute uniqueness, distinctness and entropy.

                The values should be booleans indicating the setting.
                To fully turn off statistics computation pass `statistics_config=False`.
                Defaults to `None` and will compute only descriptive statistics.
            event_time:
                Optionally, provide the name of the feature containing the event time for the features in this feature group.
                If event_time is set the feature group can be used for point-in-time joins.

                Note: Event time data type restriction
                    The supported data types for the event time column are: `timestamp`, `date` and `bigint`.

            topic_name:
                Optionally, define the name of the topic used for data ingestion.
                If left undefined it defaults to using project topic.
            notification_topic_name:
                Optionally, define the name of the topic used for sending notifications when entries are inserted or updated on the online feature store.
                If left undefined no notifications are sent.
            expectation_suite:
                Optionally, attach an expectation suite to the feature group which dataframes should be validated against upon insertion.
            data_source:
                The data source specifying the location of the data.
                Overrides the path and query arguments when specified.
            online_enabled:
                Define whether it should be possible to sync the feature group to the online feature store for low latency access.
            ttl:
                Optional time-to-live duration for features in this group.

                Can be specified as:

                - An integer or float representing seconds
                - A timedelta object

                This ttl value is added to the event time of the feature group and when the system time exceeds the event time + ttl, the entries will be automatically removed.
                The system time zone is in UTC.
                By default no TTL is set.
            ttl_enabled:
                Optionally, enable TTL for this feature group.
                Defaults to True if ttl is set.

        Returns:
            The external feature group metadata object.
        """
        if not data_source:
            data_source = ds.DataSource(query=query, path=path)
        feature_group_object = feature_group.ExternalFeatureGroup(
            name=name,
            data_format=data_format,
            options=options or {},
            storage_connector=storage_connector,
            version=version,
            description=description,
            primary_key=primary_key or [],
            foreign_key=foreign_key or [],
            featurestore_id=self._id,
            featurestore_name=self._name,
            features=features or [],
            statistics_config=statistics_config,
            event_time=event_time,
            expectation_suite=expectation_suite,
            topic_name=topic_name,
            notification_topic_name=notification_topic_name,
            data_source=data_source,
            online_enabled=online_enabled,
            ttl=ttl,
            ttl_enabled=ttl_enabled,
        )
        feature_group_object.feature_store = self
        return feature_group_object

    @usage.method_logger
    def create_external_feature_group(
        self,
        name: str,
        storage_connector: storage_connector.StorageConnector,
        query: str | None = None,
        data_format: str | None = None,
        path: str | None = "",
        options: dict[str, str] | None = None,
        version: int | None = None,
        description: str | None = "",
        primary_key: list[str] | None = None,
        foreign_key: list[str] | None = None,
        embedding_index: EmbeddingIndex | None = None,
        features: list[feature.Feature] | None = None,
        statistics_config: StatisticsConfig | bool | dict | None = None,
        event_time: str | None = None,
        expectation_suite: expectation_suite.ExpectationSuite
        | TypeVar("great_expectations.core.ExpectationSuite")
        | None = None,
        online_enabled: bool = False,
        topic_name: str | None = None,
        notification_topic_name: str | None = None,
        online_config: OnlineConfig | dict[str, Any] | None = None,
        data_source: ds.DataSource | dict[str, Any] | None = None,
        ttl: float | timedelta | None = None,
        ttl_enabled: bool | None = None,
        online_disk: bool | None = None,
    ) -> feature_group.ExternalFeatureGroup:
        """Create an external feature group metadata object.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            external_fg = fs.create_external_feature_group(
                name="sales",
                version=1,
                description="Physical shop sales features",
                query=query,
                storage_connector=connector,
                primary_key=['ss_store_sk'],
                event_time='sale_date',
                ttl=timedelta(days=30),
            )
            ```

        Note: Lazy
            This method is lazy and does not persist any metadata in the
            feature store on its own.
            To persist the feature group metadata in the feature store, call the `save()` method.

        You can enable online storage for external feature groups, however, the sync from the external storage to Hopsworks online storage needs to be done manually:

        ```python
        external_fg = fs.create_external_feature_group(
            name="sales",
            version=1,
            description="Physical shop sales features",
            query=query,
            storage_connector=connector,
            primary_key=['ss_store_sk'],
            event_time='sale_date',
            online_enabled=True,
            online_config={'online_comments': ['NDB_TABLE=READ_BACKUP=1']},
            online_disk=True, # Online data will be stored on disk instead of in memory
            ttl=timedelta(days=30),
        )
        external_fg.save()

        # read from external storage and filter data to sync to online
        df = external_fg.read().filter(external_fg.customer_status == "active")

        # insert to online storage
        external_fg.insert(df)
        ```

        Parameters:
            name: Name of the external feature group to create.
            storage_connector: The storage connector used to establish connectivity with the data source.
            query:
                A string containing a SQL query valid for the target data source.
                The query will be used to pull data from the data sources when the feature group is used.
            data_format:
                If the external feature groups refers to a directory with data, the data format to use when reading it.
            path:
                The location within the scope of the storage connector, from where to read the data for the external feature group.
            options:
                Additional options to be used by the engine when reading data from the specified storage connector.
                For example, `{"header": True}` when reading CSV files with column names in the first row.
            version:
                Version of the external feature group to retrieve, defaults to `None` and will create the feature group with incremented version from the last version in the feature store.
            description:
                A string describing the contents of the external feature group to improve discoverability for Data Scientists.
            primary_key:
                A list of feature names to be used as primary key for the feature group.
                This primary key can be a composite key of multiple features and will be used as joining key, if not specified otherwise.
                Defaults to empty list `[]`, and the feature group won't have any primary key.
            foreign_key:
                A list of feature names to be used as foreign key for the feature group.
                Foreign key is referencing the primary key of another feature group and can be used as joining key.
                Defaults to empty list `[]`, and the feature group won't have any foreign key.
            features:
                Optionally, define the schema of the external feature group manually as a list of `Feature` objects.
                Defaults to empty list `[]` and will use the schema information of the DataFrame resulting by executing the provided query against the data source.
            statistics_config:
                A configuration object, or a dictionary with keys:

                - `"enabled"` to generally enable descriptive statistics computation for this external feature group,
                - `"correlations"` to turn on feature correlation
                computation,
                - `"histograms"` to compute feature value frequencies, and
                - `"exact_uniqueness"` to compute uniqueness, distinctness and entropy.

                The values should be booleans indicating the setting. To fully turn off statistics computation pass `statistics_config=False`.
                Defaults to `None` and will compute only descriptive statistics.
            event_time:
                Optionally, provide the name of the feature containing the event time for the features in this feature group.
                If event_time is set the feature group can be used for point-in-time joins.

                Note: Event time data type restriction
                    The supported data types for the event time column are: `timestamp`, `date` and `bigint`.

            online_enabled:
                Define whether it should be possible to sync the feature group to the online feature store for low latency access.
            expectation_suite:
                Optionally, attach an expectation suite to the feature group which dataframes should be validated against upon insertion.
            topic_name:
                Optionally, define the name of the topic used for data ingestion.
                If left undefined it defaults to using project topic.
            notification_topic_name:
                Optionally, define the name of the topic used for sending notifications when entries are inserted or updated on the online feature store.
                If left undefined no notifications are sent.
            online_config:
                Optionally, define configuration which is used to configure online table.
            data_source:
                The data source specifying the location of the data.
                Overrides the path and query arguments when specified.
            ttl:
                Optional time-to-live duration for features in this group.

                Can be specified as:

                - An integer or float representing seconds
                - A timedelta object

                This ttl value is added to the event time of the feature group and when the system time exceeds the event time + ttl, the entries will be automatically removed.
                The system time zone is in UTC.
                By default no TTL is set.
            ttl_enabled:
                Optionally, enable TTL for this feature group.
                Defaults to True if ttl is set.
            online_disk:
                Optionally, specify online data storage for this feature group.
                When set to True data will be stored on disk, instead of in memory.
                Overrides online_config.table_space.
                Defaults to using cluster wide configuration 'featurestore_online_tablespace' to identify tablespace for disk storage.

        Returns:
            The external feature group metadata object.
        """
        if not data_source:
            data_source = ds.DataSource(query=query, path=path)
        feature_group_object = feature_group.ExternalFeatureGroup(
            name=name,
            data_format=data_format,
            options=options or {},
            storage_connector=storage_connector,
            version=version,
            description=description,
            primary_key=primary_key or [],
            foreign_key=foreign_key or [],
            embedding_index=embedding_index,
            featurestore_id=self._id,
            featurestore_name=self._name,
            features=features or [],
            statistics_config=statistics_config,
            event_time=event_time,
            expectation_suite=expectation_suite,
            online_enabled=online_enabled,
            topic_name=topic_name,
            notification_topic_name=notification_topic_name,
            online_config=online_config,
            data_source=data_source,
            ttl=ttl,
            ttl_enabled=ttl_enabled,
            online_disk=online_disk,
        )
        feature_group_object.feature_store = self
        return feature_group_object

    @usage.method_logger
    def get_or_create_spine_group(
        self,
        name: str,
        version: int | None = None,
        description: str | None = "",
        primary_key: list[str] | None = None,
        foreign_key: list[str] | None = None,
        event_time: str | None = None,
        features: list[feature.Feature] | None = None,
        dataframe: pd.DataFrame
        | TypeVar("pyspark.sql.DataFrame")
        | TypeVar("pyspark.RDD")
        | np.ndarray
        | list[list] = None,
    ) -> feature_group.SpineGroup:
        """Create a spine group metadata object.

        Instead of using a feature group to save a label/prediction target, you can use a spine together with a dataframe containing the labels.
        A Spine is essentially a metadata object similar to a feature group, however, the data is not materialized in the feature store.
        It only containes the needed metadata such as the relevant event time column and primary key columns to perform point-in-time correct joins.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            spine_df = pd.Dataframe()

            spine_group = fs.get_or_create_spine_group(
                name="sales",
                version=1,
                description="Physical shop sales features",
                primary_key=['ss_store_sk'],
                event_time='sale_date',
                dataframe=spine_df,
            )
            ```

        Note that you can inspect the dataframe in the spine group, or replace the dataframe:

        ```python
        spine_group.dataframe.show()

        spine_group.dataframe = new_df
        ```

        The spine can then be used to construct queries, with only one speciality:

        Note:
            Spines can only be used on the left side of a feature join, as this is the base set of entities for which features are to be fetched and the left side of the join determines the event timestamps to compare against.

        **If you want to use the query for a feature view to be used for online serving, you can only select the label or target feature from the spine.**
        For the online lookup, the label is not required, therefore it is important to only select label from the left feature group, so that we don't need to provide a spine for online serving.

        These queries can then be used to create feature views.
        Since the dataframe contained in the spine is not being materialized, every time you use a feature view created with spine to read data you will have to provide a dataframe with the same structure again.

        For example, to generate training data:

        ```python
        X_train, X_test, y_train, y_test = feature_view_spine.train_test_split(0.2, spine=training_data_entities)
        ```

        Or to get batches of fresh data for batch scoring:

        ```python
        feature_view_spine.get_batch_data(spine=scoring_entities_df).show()
        ```

        Here you have the chance to pass a different set of entities to generate the training dataset.

        Sometimes it might be handy to create a feature view with a regular feature group containing the label, but then at serving time to use a spine in order to fetch features for example only for a small set of primary key values.
        To do this, you can pass the spine group instead of a dataframe.
        Just make sure it contains the needed primary key, event time and label column.

        ```python
        feature_view.get_batch_data(spine=spine_group)
        ```

        Parameters:
            name: Name of the spine group to create.
            version: Version of the spine group to retrieve, defaults to `None` and will create the spine group with incremented version from the last version in the feature store.
            description: A string describing the contents of the spine group to improve discoverability for Data Scientists.
            primary_key:
                A list of feature names to be used as primary key for the spine group.
                This primary key can be a composite key of multiple features and will be used as joining key, if not specified otherwise.
                Defaults to empty list `[]`, and the spine group won't have any primary key.
            foreign_key:
                A list of feature names to be used as foreign key for the feature group.
                Foreign key is referencing the primary key of another feature group and can be used as joining key.
                Defaults to empty list `[]`, and the feature group won't have any foreign key.
            event_time:
                Optionally, provide the name of the feature containing the event time for the features in this spine group.
                If event_time is set the spine group can be used for point-in-time joins.
            features:
                Optionally, define the schema of the spine group manually as a list of `Feature` objects.
                Defaults to empty list `[]` and will use the schema information of the DataFrame resulting by executing the provided query against the data source.

                Note: Event time data type restriction
                    The supported data types for the event time column are: `timestamp`, `date` and `bigint`.

            dataframe: Spine dataframe with primary key, event time and label column to use for point in time join when fetching features.

        Returns:
            The spine group metadata object.
        """
        spine = self._feature_group_api.get(self.id, name, version)
        if spine:
            spine.dataframe = dataframe
            spine.feature_store = self
            return spine
        spine = feature_group.SpineGroup(
            name=name,
            version=version,
            description=description,
            primary_key=primary_key or [],
            foreign_key=foreign_key or [],
            event_time=event_time,
            features=features or [],
            dataframe=dataframe,
            featurestore_id=self._id,
            featurestore_name=self._name,
        )
        spine.feature_store = self
        return spine._save()

    def create_training_dataset(
        self,
        name: str,
        version: int | None = None,
        description: str | None = "",
        data_format: str | None = "tfrecords",
        coalesce: bool | None = False,
        storage_connector: storage_connector.StorageConnector | None = None,
        splits: dict[str, float] | None = None,
        location: str | None = "",
        seed: int | None = None,
        statistics_config: StatisticsConfig | bool | dict | None = None,
        label: list[str] | None = None,
        transformation_functions: dict[str, TransformationFunction] | None = None,
        train_split: str = None,
        tags: tag.Tag | dict[str, Any] | list[tag.Tag | dict[str, Any]] | None = None,
    ) -> training_dataset.TrainingDataset:
        """Create a training dataset metadata object.

        Warning: Deprecated
            `TrainingDataset` is deprecated, use `FeatureView` instead.
            From version 3.0 training datasets created with this API are not visibile in the API anymore.

        Note: Lazy
            This method is lazy and does not persist any metadata or feature data in the feature store on its own.
            To materialize the training dataset and save feature data along the metadata in the feature store, call the `save()` method with a `DataFrame` or `Query`.

        Info: Data Formats
            The feature store currently supports the following data formats for
            training datasets:

            1. tfrecord
            2. csv
            3. tsv
            4. parquet
            5. avro
            6. orc

            Currently not supported petastorm, hdf5 and npy file formats.

        Parameters:
            name: Name of the training dataset to create.
            version: Version of the training dataset to retrieve, defaults to `None` and will create the training dataset with incremented version from the last version in the feature store.
            description: A string describing the contents of the training dataset to improve discoverability for Data Scientists.
            data_format: The data format used to save the training dataset.
            coalesce:
                If true the training dataset data will be coalesced into a single partition before writing.
                The resulting training dataset will be a single file per split.
            storage_connector:
                Storage connector defining the sink location for the training dataset, defaults to `None`, and materializes training dataset on HopsFS.
            splits:
                A dictionary defining training dataset splits to be created.
                Keys in the dictionary define the name of the split as `str`, values represent percentage of samples in the split as `float`.
                Currently, only random splits are supported.
                Defaults to empty dict`{}`, creating only a single training dataset without splits.
            location:
                Path to complement the sink storage connector with, e.g., if the storage connector points to an S3 bucket, this path can be used to define a sub-directory inside the bucket to place the training dataset.
                Defaults to `""`, saving the training dataset at the root defined by the storage connector.
            seed: Optionally, define a seed to create the random splits with, in order to guarantee reproducability.
            statistics_config:
                A configuration object, or a dictionary with keys:
                - `"enabled"` to generally enable descriptive statistics computation for this feature group,
                - `"correlations"` to turn on feature correlation computation, and
                - `"histograms"` to compute feature value frequencies.

                The values should be booleans indicating the setting.
                To fully turn off statistics computation pass `statistics_config=False`.
                Defaults to `None` and will compute only descriptive statistics.
            label:
                A list of feature names constituting the prediction label/feature of the training dataset.
                When replaying a `Query` during model inference, the label features can be omitted from the feature vector retrieval.
                Defaults to `[]`, no label.
            transformation_functions:
                A dictionary mapping transformation functions to the features they should be applied to before writing out the training data and at inference time.
                Defaults to `{}`, no transformations.
            train_split:
                If `splits` is set, provide the name of the split that is going to be used for training.
                The statistics of this split will be used for transformation functions if necessary.
            tags:
                Optionally, define tags for the training dataset. Tags can be provided as:
                - A single Tag object
                - A dictionary with 'name' and 'value' keys (e.g., {"name": "tag1", "value": "value1"})
                - A list of Tag objects
                - A list of dictionaries with 'name' and 'value' keys
                Tags will be attached to the training dataset after it is saved. Defaults to None.

        Returns:
            The training dataset metadata object.
        """
        normalized_tags = self._normalize_tags(tags)

        return training_dataset.TrainingDataset(
            name=name,
            version=version,
            description=description,
            data_format=data_format,
            storage_connector=storage_connector,
            location=location,
            featurestore_id=self._id,
            splits=splits or {},
            seed=seed,
            statistics_config=statistics_config,
            label=label or [],
            coalesce=coalesce,
            transformation_functions=transformation_functions or {},
            train_split=train_split,
            tags=normalized_tags,
        )

    @usage.method_logger
    def create_transformation_function(
        self,
        transformation_function: HopsworksUdf,
        version: int | None = None,
    ) -> TransformationFunction:
        """Create a transformation function metadata object.

        Example:
            ```python
            # define the transformation function as a Hopsworks's UDF
            @udf(int)
            def plus_one(value):
                return value + 1

            # create transformation function
            plus_one_meta = fs.create_transformation_function(
                    transformation_function=plus_one,
                    version=1
                )

            # persist transformation function in backend
            plus_one_meta.save()
            ```

        Note: Lazy
            This method is lazy and does not persist the transformation function in the feature store on its own.
            To materialize the transformation function and save call the `save()` method of the transformation function metadata object.

        Parameters:
            transformation_function: Hopsworks UDF.

        Returns:
            The TransformationFunction metadata object.
        """
        return TransformationFunction(
            featurestore_id=self._id,
            hopsworks_udf=transformation_function,
            version=version,
        )

    @usage.method_logger
    def get_transformation_function(
        self,
        name: str,
        version: int | None = None,
    ) -> TransformationFunction:
        """Get  transformation function metadata object.

        Example: Get transformation function by name
            This will default to version 1.

            ```python
            # get feature store instance
            fs = ...

            # get transformation function metadata object
            plus_one_fn = fs.get_transformation_function(name="plus_one")
            ```

        Example: Get built-in transformation function min max scaler
            ```python
            # get feature store instance
            fs = ...

            # get transformation function metadata object
            min_max_scaler_fn = fs.get_transformation_function(name="min_max_scaler")
            ```

        Example: Get transformation function by name and version
            ```python
            # get feature store instance
            fs = ...

            # get transformation function metadata object
            min_max_scaler = fs.get_transformation_function(name="min_max_scaler", version=2)
            ```

        You can define in the feature view transformation functions as dict, where key is feature name and value is online transformation function instance.
        Then the transformation functions are applied when you read training data, get batch data, or get feature vector(s).

        Example: Attach transformation functions to the feature view
            ```python
            # get feature store instance
            fs = ...

            # define query object
            query = ...

            # get transformation function metadata object
            min_max_scaler = fs.get_transformation_function(name="min_max_scaler", version=1)

            # attach transformation functions
            feature_view = fs.create_feature_view(
                name='feature_view_name',
                query=query,
                labels=["target_column"],
                transformation_functions=[min_max_scaler("feature1")]
            )
            ```

        Built-in transformation functions are attached in the same way.
        The only difference is that it will compute the necessary statistics for the specific function in the background.
        For example min and max values for `min_max_scaler`; mean and standard deviation for `standard_scaler` etc.

        Example: Attach built-in transformation functions to the feature view
            ```python
            # get feature store instance
            fs = ...

            # define query object
            query = ...

            # retrieve transformation functions
            min_max_scaler = fs.get_transformation_function(name="min_max_scaler")
            standard_scaler = fs.get_transformation_function(name="standard_scaler")
            robust_scaler = fs.get_transformation_function(name="robust_scaler")
            label_encoder = fs.get_transformation_function(name="label_encoder")

            # attach built-in transformation functions while creating feature view
            feature_view = fs.create_feature_view(
                name='transactions_view',
                query=query,
                labels=["fraud_label"],
                transformation_functions = [
                    label_encoder("category_column"),
                    robust_scaler("weight"),
                    min_max_scaler("age"),
                    standard_scaler("salary")
                ]
            )
            ```

        Parameters:
            name: Name of transformation function.
            version:
                Version of transformation function.
                Optional, if not provided all functions that match to provided name will be retrieved.

        Returns:
            The TransformationFunction metadata object.
        """
        return self._transformation_function_engine.get_transformation_fn(name, version)

    @usage.method_logger
    def get_transformation_functions(self) -> list[TransformationFunction]:
        """Get  all transformation functions metadata objects.

        Example: Get all transformation functions
            ```python
            # get feature store instance
            fs = ...

            # get all transformation functions
            list_transformation_fns = fs.get_transformation_functions()
            ```

        Returns:
            List of transformation function instances.
        """
        return self._transformation_function_engine.get_transformation_fns()

    @usage.method_logger
    def create_feature_view(
        self,
        name: str,
        query: Query,
        version: int | None = None,
        description: str | None = "",
        labels: list[str] | None = None,
        inference_helper_columns: list[str] | None = None,
        training_helper_columns: list[str] | None = None,
        transformation_functions: list[TransformationFunction | HopsworksUdf]
        | None = None,
        logging_enabled: bool | None = False,
        extra_log_columns: list[feature.Feature] | list[dict[str, str]] | None = None,
        tags: tag.Tag | dict[str, Any] | list[tag.Tag | dict[str, Any]] | None = None,
    ) -> feature_view.FeatureView:
        """Create a feature view metadata object and saved it to hopsworks.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            # get the feature group instances
            fg1 = fs.get_or_create_feature_group(...)
            fg2 = fs.get_or_create_feature_group(...)

            # construct the query
            query = fg1.select_all().join(fg2.select_all())

            # define the transformation function as a Hopsworks's UDF
            @udf(int)
            def plus_one(value):
                return value + 1

            # construct list of "transformation functions" on features
            transformation_functions = [plus_one("feature1"), plus_one("feature1"))]

            feature_view = fs.create_feature_view(
                name='air_quality_fv',
                version=1,
                transformation_functions=transformation_functions,
                query=query
            )
            ```

        Example:
            ```python
            # get feature store instance
            fs = ...

            # define query object
            query = ...

            # define list of transformation functions
            mapping_transformers = ...

            # create feature view
            feature_view = fs.create_feature_view(
                name='feature_view_name',
                version=1,
                transformation_functions=mapping_transformers,
                query=query
            )
            ```

        Warning:
            `as_of` argument in the `Query` will be ignored because feature view does not support time travel query.

        Parameters:
            name: Name of the feature view to create.
            query: Feature store `Query`.
            version: Version of the feature view to create, defaults to `None` and will create the feature view with incremented version from the last version in the feature store.
            description: A string describing the contents of the feature view to improve discoverability for Data Scientists.
            labels:
                A list of feature names constituting the prediction label/feature of the feature view.
                When replaying a `Query` during model inference, the label features can be omitted from the feature vector retrieval.
                Defaults to `[]`, no label.
            inference_helper_columns:
                A list of feature names that are not used in training the model itself but can be used during batch or online inference for extra information.
                Inference helper column name(s) must be part of the `Query` object.
                If inference helper column name(s) belong to feature group that is part of a `Join` with `prefix` defined, then this prefix needs to be prepended to the original column name when defining `inference_helper_columns` list.
                When replaying a `Query` during model inference, the inference helper columns optionally can be omitted during batch (`get_batch_data`) and will be omitted during online  inference (`get_feature_vector(s)`).
                To get inference helper column(s) during online inference use `get_inference_helper(s)` method.
                Defaults to `[], no helper columns.
            training_helper_columns:
                A list of feature names that are not the part of the model schema itself but can be used during training as a helper for extra information.
                Training helper column name(s) must be part of the `Query` object.
                If training helper column name(s) belong to feature group that is part of a `Join` with `prefix` defined, then this prefix needs to prepended to the original column name when defining `training_helper_columns` list.
                When replaying a `Query` during model inference, the training helper columns will be omitted during both batch and online inference.
                Training helper columns can be optionally fetched with training data.
                For more details see documentation for feature view's get training data methods.
                Defaults to `[]`, no training helper columns.
            transformation_functions:
                Model Dependent Transformation functions attached to the feature view.
                It can be a list of list of user defined functions defined using the hopsworks `@udf` decorator.
                Defaults to `None`, no transformations.
            logging_enabled: If true, enable feature logging for the feature view.
            extra_log_columns:
                Extra columns to be logged in addition to the features used in the feature view.
                It can be a list of Feature objects or list a dictionaries that contains the the name and type of the columns as keys.
                Defaults to `None`, no extra log columns. Setting this argument implicitly enables feature logging.
            tags:
                Optionally, define tags for the feature view. Tags can be provided as:
                - A single Tag object
                - A dictionary with 'name' and 'value' keys (e.g., {"name": "tag1", "value": "value1"})
                - A list of Tag objects
                - A list of dictionaries with 'name' and 'value' keys
                Tags will be attached to the feature view after it is saved. Defaults to None.

        Returns:
            The feature view metadata object.
        """
        normalized_tags = self._normalize_tags(tags)

        feat_view = feature_view.FeatureView(
            name=name,
            query=query,
            featurestore_id=self._id,
            version=version,
            description=description,
            labels=labels or [],
            inference_helper_columns=inference_helper_columns or [],
            training_helper_columns=training_helper_columns or [],
            transformation_functions=transformation_functions or {},
            featurestore_name=self._name,
            logging_enabled=logging_enabled,
            extra_log_columns=extra_log_columns,
            tags=normalized_tags,
        )
        return self._feature_view_engine.save(feat_view)

    @usage.method_logger
    def get_or_create_feature_view(
        self,
        name: str,
        query: Query,
        version: int,
        description: str | None = "",
        labels: list[str] | None = None,
        inference_helper_columns: list[str] | None = None,
        training_helper_columns: list[str] | None = None,
        transformation_functions: dict[str, TransformationFunction] | None = None,
        logging_enabled: bool | None = False,
        extra_log_columns: list[feature.Feature] | list[dict[str, str]] | None = None,
    ) -> feature_view.FeatureView:
        """Get feature view metadata object or create a new one if it doesn't exist.

        This method doesn't update existing feature view metadata object.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            feature_view = fs.get_or_create_feature_view(
                name='bitcoin_feature_view',
                version=1,
                transformation_functions=transformation_functions,
                query=query
            )
            ```

        Parameters:
            name: Name of the feature view to create.
            query: Feature store `Query`.
            version: Version of the feature view to create.
            description: A string describing the contents of the feature view to improve discoverability for Data Scientists.
            labels:
                A list of feature names constituting the prediction label/feature of the feature view.
                When replaying a `Query` during model inference, the label features can be omitted from the feature vector retrieval.
                Defaults to `[]`, no label.
            inference_helper_columns:
                A list of feature names that are not used in training the model itself but can be used during batch or online inference for extra information.
                Inference helper column name(s) must be part of the `Query` object.
                If inference helper column name(s) belong to feature group that is part of a `Join` with `prefix` defined, then this prefix needs to be prepended to the original column name when defining `inference_helper_columns` list.
                When replaying a `Query` during model inference, the inference helper columns optionally can be omitted during batch (`get_batch_data`) and will be omitted during online  inference (`get_feature_vector(s)`).
                To get inference helper column(s) during online inference use `get_inference_helper(s)` method.
                Defaults to `[], no helper columns.
            training_helper_columns:
                A list of feature names that are not the part of the model schema itself but can be used during training as a helper for extra information.
                Training helper column name(s) must be part of the `Query` object.
                If training helper column name(s) belong to feature group that is part of a `Join` with `prefix` defined, then this prefix needs to prepended to the original column name when defining `training_helper_columns` list.
                When replaying a `Query` during model inference, the training helper columns will be omitted during both batch and online inference.
                Training helper columns can be optionally fetched with training data.
                For more details see documentation for feature view's get training data methods.
                Defaults to `[]`, no training helper columns.
            transformation_functions:
                Model Dependent Transformation functions attached to the feature view.
                It can be a list of list of user defined functions defined using the hopsworks `@udf` decorator.
                Defaults to `None`, no transformations.
            logging_enabled: If true, enable feature logging for the feature view.
            extra_log_columns:
                Extra columns to be logged in addition to the features used in the feature view.
                It can be a list of Feature objects or list a dictionaries that contains the the name and type of the columns as keys.
                Defaults to `None`, no extra log columns.
                Setting this argument implicitly enables feature logging.

        Returns:
            The feature view metadata object.
        """
        fv_object = self._feature_view_engine.get(name, version)
        if not fv_object:
            fv_object = self.create_feature_view(
                name=name,
                query=query,
                version=version,
                description=description,
                labels=labels or [],
                inference_helper_columns=inference_helper_columns or [],
                training_helper_columns=training_helper_columns or [],
                transformation_functions=transformation_functions or [],
                logging_enabled=logging_enabled,
                extra_log_columns=extra_log_columns,
            )
        return fv_object

    @usage.method_logger
    def get_feature_view(
        self, name: str, version: int = None
    ) -> feature_view.FeatureView:
        """Get a feature view entity from the feature store.

        Getting a feature view from the Feature Store means getting its metadata.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get feature view instance
            feature_view = fs.get_feature_view(
                name='feature_view_name',
                version=1
            )
            ```

        Parameters:
            name: Name of the feature view to get.
            version: Version of the feature view to retrieve, defaults to `None` and will return the `version=1`.

        Returns:
            The feature view metadata object or `None` if it does not exist.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        if version is None:
            warnings.warn(
                f"No version provided for getting feature view `{name}`, defaulting to `{self.DEFAULT_VERSION}`.",
                util.VersionWarning,
                stacklevel=1,
            )
            version = self.DEFAULT_VERSION
        feature_view_object = self._feature_view_engine.get(name, version)
        if feature_view_object:
            util.check_missing_mandatory_tags(
                feature_view_object.missing_mandatory_tags
            )
        return feature_view_object

    @usage.method_logger
    def get_feature_views(self, name: str) -> list[feature_view.FeatureView]:
        """Get a list of all versions of a feature view entity from the feature store.

        Getting a feature view from the Feature Store means getting its metadata.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get a list of all versions of a feature view
            feature_view = fs.get_feature_views(
                name='feature_view_name'
            )
            ```

        Parameters:
            name: Name of the feature view to get.

        Returns:
            List of feature view metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._feature_view_engine.get(name)

    def _disable_hopsworks_feature_query_service_client(self):
        """Disable Hopsworks feature query service for the current session. This behaviour is not persisted on reset."""
        from hsfs.core import arrow_flight_client

        arrow_flight_client._disable_feature_query_service_client()

    def _reset_hopsworks_feature_query_service_client(self):
        """Reset Hopsworks feature query service for the current session."""
        from hsfs.core import arrow_flight_client

        arrow_flight_client.close()
        arrow_flight_client.get_instance()

    def create_chart(
        self, title: str, description: str, url: str, job_id: int | None = None
    ) -> None:
        """Create a chart in the feature store.

        Registers an HTML file as a chart in Hopsworks.
        This enables it to be used in a [`Dashboard`][hsfs.core.dashboard.Dashboard].

        Each chart with a set `job_id` has a refresh button which triggers the job and redraws the chart once the job finishes.
        You can use this job to conviniently extract and prepare the data from Hopsworks Feature Store using its Python API.
        Once the data is acquired, it can be put into JSON to simplify the Javascript code in the HTML.

        Note: Jobless charts
            Although charts can be created without a data preparation job, such charts are not suited to visualize data stored in Hopsworks.
            Jobless charts can be useful, for example, in case you want to display data which is already available in JSON via a REST API of an external service, or if the chart is completely static.
            Jobless charts do not have a refresh button attached to them.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # create a chart
            fs.create_chart(
                title="My Chart",
                description="This is my chart description",
                url="/Resources/chart.html"
            )
            ```

        Arguments:
            title: Title of the chart.
            description: Description of the chart.
            url: URL where the chart is hosted or can be accessed.
            job_id: ID of the job that prepares the data to be displayed in the chart.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        chart = Chart(
            title=title,
            description=description,
            url=url,
            job=Job(id=job_id) if job_id else None,
        )
        return ChartApi().create_chart(chart)

    def get_charts(self) -> list[Chart]:
        """Get all charts in the feature store.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get all charts
            charts = fs.get_charts()
            ```

        Returns:
            List of chart metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return ChartApi().get_charts()

    def get_chart(self, chart_id: int) -> Chart:
        """Get a chart by its ID.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get a specific chart
            chart = fs.get_chart(chart_id=123)
            ```

        Arguments:
            chart_id: ID of the chart to retrieve.

        Returns:
            The chart metadata object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return ChartApi().get_chart(chart_id)

    def create_dashboard(self, name: str, charts: list[Chart] | None = None) -> None:
        """Create a dashboard in the feature store.

        Example:
            ```python
            # get feature store instance
            fs = ...

            chart = fs.get_chart(chart_id=321)
            chart.width = 12
            chart.height = 8
            chart.x = 0
            chart.y = 0

            # create a dashboard
            fs.create_dashboard(
                name="My Dashboard",
                charts=[chart]  # optional
            )
            ```

        Arguments:
            name: Name of the dashboard.
            charts: List of charts to include in the dashboard.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        dashboard = Dashboard(
            name=name,
            charts=charts,
        )
        return DashboardApi().create_dashboard(dashboard)

    def get_dashboards(self) -> list[Dashboard]:
        """Get all dashboards in the feature store.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get all dashboards
            dashboards = fs.get_dashboards()
            ```

        Returns:
            List of dashboard metadata objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return DashboardApi().get_dashboards()

    def get_dashboard(self, dashboard_id: int) -> Dashboard:
        """Get a dashboard by its ID.

        Example:
            ```python
            # get feature store instance
            fs = ...

            # get a specific dashboard
            dashboard = fs.get_dashboard(dashboard_id=123)
            ```

        Arguments:
            dashboard_id: ID of the dashboard to retrieve.

        Returns:
            The dashboard metadata object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return DashboardApi().get_dashboard(dashboard_id)

    @property
    def id(self) -> int:
        """Id of the feature store."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the feature store."""
        return self._name

    @property
    def project_name(self) -> str:
        """Name of the project in which the feature store is located."""
        return self._project_name

    @property
    def project_id(self) -> int:
        """Id of the project in which the feature store is located."""
        return self._project_id

    @property
    def online_featurestore_name(self) -> str | None:
        """Name of the online feature store database."""
        return self._online_feature_store_name

    @property
    def online_enabled(self) -> bool:
        """Indicator whether online feature store is enabled."""
        return self._online_enabled

    @property
    def offline_featurestore_name(self) -> str:
        """Name of the offline feature store database."""
        return self._offline_feature_store_name

    @usage.method_logger
    def search(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | search_api.TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> search_api.FeaturestoreSearchResult:
        """Search for feature groups, feature views, training datasets and features.

        Parameters:
           search_term: the term to search for.
           keyword_filter: filter results by keywords. Can be a single string or an array of strings.
           tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
           offset: the number of results to skip (default is 0).
           limit: the number of search results to return (default is 100).
           global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
           `FeaturestoreSearchResult`: The search results containing lists of metadata objects for feature groups, feature views, training datasets, and features.

        Raises:
           `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        fs = project.get_feature_store()

        # Simple search
        result = fs.search("search-term")

        # Access results
        for fg_meta in result.feature_groups:
           print(f"Feature Group: {fg_meta.name} v{fg_meta.version}")
           print(f"Description: {fg_meta.description}")
           print(f"Highlights: {fg_meta.highlights}")

           # Get the same FeatureGroup object as returned by featurestore.get_feature_group
           fg = fg_meta.get()

        # Search with a single keyword (string)
        result = fs.search("search-term", keyword_filter="ml")

        # Search with multiple keywords (array of strings)
        result = fs.search("search-term", keyword_filter=["ml", "production"])

        # Search with tag filter as a single dictionary
        result = fs.search(
           "search-term",
           tag_filter={"name": "tag1", "key": "environment", "value": "production"}
        )

        # Search with tag filter as an array of dictionaries
        result = fs.search(
           "search-term",
           tag_filter=[
               {"name": "tag1", "key": "environment", "value": "production"},
               {"name": "tag2", "key": "version", "value": "v1.0"}
           ]
        )

        # Search with TagSearchFilter objects
        from hsfs.core.search_api import TagSearchFilter
        tags = [
           TagSearchFilter(name="tag1", key="environment", value="production"),
           TagSearchFilter(name="tag2", key="version", value="v1.0")
        ]
        result = fs.search("search-term", tag_filter=tags)

        # Search with both keyword_filter and tag_filter
        result = fs.search(
           "search-term",
           keyword_filter=["ml", "production"],
           tag_filter=tags
        )
        ```
        """
        return self._search_api.feature_store(
            search_term=search_term,
            tag_filter=tag_filter,
            keyword_filter=keyword_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )

    @usage.method_logger
    def search_feature_groups(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | search_api.TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[search_api.FeatureGroupSearchResult]:
        """Search for feature groups only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for feature groups matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        fs = project.get_feature_store()

        # Search for feature groups
        fg_metas = fs.search_feature_groups("customer")

        for fg_meta in fg_metas:
            print(f"Feature Group: {fg_meta.name} v{fg_meta.version}")

            # Get the same FeatureGroup object as returned by featurestore.get_feature_group
            fg = fg_meta.get()
        ```
        """
        return self._search_api.feature_groups(
            search_term=search_term,
            tag_filter=tag_filter,
            keyword_filter=keyword_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )

    @usage.method_logger
    def search_feature_views(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | search_api.TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[search_api.FeatureViewSearchResult]:
        """Search for feature views only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for feature views matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        fs = project.get_feature_store()

        # Search for feature views
        fv_metas = fs.search_feature_views("customer")

        for fv_meta in fv_metas:
            print(f"Feature View: {fv_meta.name} v{fv_meta.version}")

            # Get the same FeatureView object as returned by featurestore.get_feature_view
            fv = fv_meta.get()
        ```
        """
        return self._search_api.feature_views(
            search_term=search_term,
            tag_filter=tag_filter,
            keyword_filter=keyword_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )

    @usage.method_logger
    def search_training_datasets(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | search_api.TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[search_api.TrainingDatasetSearchResult]:
        """Search for training datasets only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of metadata objects for training datasets matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        fs = project.get_feature_store()

        # Search for training datasets
        td_metas = fs.search_training_datasets("model")

        for td_meta in td_metas:
            print(f"Training Dataset: {td_meta.name} v{td_meta.version}")

            # Get the same TrainingDataset object as returned by featurestore.get_training_dataset
            td = td_meta.get()
        ```
        """
        return self._search_api.training_datasets(
            search_term=search_term,
            tag_filter=tag_filter,
            keyword_filter=keyword_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )

    @usage.method_logger
    def search_features(
        self,
        search_term: str = None,
        keyword_filter: str | list[str] | None = None,
        tag_filter: dict[str, str]
        | list[dict[str, str] | search_api.TagSearchFilter]
        | None = None,
        offset: int = 0,
        limit: int = 100,
        global_search: bool = False,
    ) -> list[search_api.FeatureSearchResult]:
        """Search for features only.

        Parameters:
            search_term: the term to search for.
            keyword_filter: filter results by keywords. Can be a single string or an array of strings.
            tag_filter: filter results by tags. Can be a single dictionary, an array of dictionaries,
               or an array of TagSearchFilter objects. Each tag filter requires: ``name`` (the tag
               schema name as defined by Hopsworks Admin), ``key`` (the property within that tag
               schema), and ``value`` (the value to match).
            offset: the number of results to skip (default is 0).
            limit: the number of search results to return (default is 100).
            global_search: By default is false - search in current project only. Set to true if you want to search over all projects

        Returns:
            `List`: A list of features matching the search criteria.

        Raises:
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request

        Example:
        ```python
        import hopsworks

        project = hopsworks.login()
        fs = project.get_feature_store()

        # Search for features
        features = fs.search_features("age")

        for feature in features:
            print(f"Feature: {feature.name}")
        ```
        """
        return self._search_api.features(
            search_term=search_term,
            tag_filter=tag_filter,
            keyword_filter=keyword_filter,
            offset=offset,
            limit=limit,
            global_search=global_search,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import time

from hopsworks_common.client.exceptions import GitException
from hopsworks_common.core import git_op_execution_api


class GitEngine:
    def __init__(self):
        self._git_op_execution_api = git_op_execution_api.GitOpExecutionApi()
        self._log = logging.getLogger(__name__)

    def execute_op_blocking(self, git_op, command: str):
        """Poll a git execution status until it reaches a terminal state.

        Parameters:
            git_op (GitOpExecution): git execution to monitor.
            command: git operation running.

        Returns:
            The final GitOpExecution object.
        """
        while git_op.success is None:
            self._log.info(f"Running command {command}, current status {git_op.state}")
            git_op = self._git_op_execution_api._get_execution(
                git_op.repository.id, git_op.id
            )
            time.sleep(5)

        if git_op.success is False:
            raise GitException(
                f"Git command failed with the message: {git_op.command_result_message}"
            )
        self._log.info(f"Git command {command} finished")

        return git_op



#
#  Copyright 2021. Logical Clocks AB
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from __future__ import annotations

import json
from typing import Any

import humps
from hsfs import util
from hsfs.constructor import prepared_statement_parameter


class ServingPreparedStatement:
    def __init__(
        self,
        feature_group_id: int | None = None,
        prepared_statement_index: int | None = None,
        prepared_statement_parameters: list[
            prepared_statement_parameter.PreparedStatementParameter
        ]
        | None = None,
        query_online: str | None = None,
        prefix: str | None = None,
        type: str | None = None,
        items: list[dict[str, Any]] | None = None,
        count: int | None = None,
        href: str | None = None,
        **kwargs,
    ) -> None:
        self._feature_group_id = feature_group_id
        self._prepared_statement_index = prepared_statement_index
        # use setter to ensure that the parameters are sorted by index
        self.prepared_statement_parameters = prepared_statement_parameters
        self._query_online = query_online
        self._prefix = prefix

    @classmethod
    def from_response_json(
        cls, json_dict: dict[str, Any]
    ) -> list[ServingPreparedStatement]:
        json_decamelized = humps.decamelize(json_dict)
        if json_decamelized["count"] == 0:
            return []
        return [cls(**pstm_dto) for pstm_dto in json_decamelized["items"]]

    def update_from_response_json(
        self, json_dict: dict[str, Any]
    ) -> ServingPreparedStatement:
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> dict[str, Any]:
        return {
            "preparedStatementIndex": self._prepared_statement_index,
            "preparedStatementParameters": self._prepared_statement_parameters,
            "queryOnline": self._query_online,
        }

    def __repr__(self) -> str:
        repr_dict = humps.decamelize(self.to_dict())
        repr_dict["feature_group_id"] = self._feature_group_id
        repr_dict["prefix"] = self._prefix
        repr_dict["prepared_statement_parameters"] = [
            pstm_param.__repr__() for pstm_param in self._prepared_statement_parameters
        ]
        return json.dumps(repr_dict, sort_keys=True, indent=4)

    @property
    def feature_group_id(self) -> int | None:
        return self._feature_group_id

    @property
    def prepared_statement_index(self) -> int | None:
        return self._prepared_statement_index

    @property
    def prepared_statement_parameters(
        self,
    ) -> list[prepared_statement_parameter.PreparedStatementParameter] | None:
        return self._prepared_statement_parameters

    @property
    def query_online(self) -> str | None:
        return self._query_online

    @property
    def prefix(self) -> str | None:
        return self._prefix

    @feature_group_id.setter
    def feature_group_id(self, feature_group_id: int | None) -> None:
        self._feature_group_id = feature_group_id

    @prepared_statement_index.setter
    def prepared_statement_index(self, prepared_statement_index: int | None) -> None:
        self._prepared_statement_index = prepared_statement_index

    @prepared_statement_parameters.setter
    def prepared_statement_parameters(
        self,
        prepared_statement_parameters: list[
            prepared_statement_parameter.PreparedStatementParameter
        ]
        | list[dict[str, Any]],
    ) -> None:
        if isinstance(prepared_statement_parameters[0], dict):
            prepared_statement_parameters = [
                prepared_statement_parameter.PreparedStatementParameter.from_response_json(
                    pstm_param
                )
                for pstm_param in prepared_statement_parameters
            ]

        self._prepared_statement_parameters = sorted(
            prepared_statement_parameters, key=lambda x: x.index
        )

    @query_online.setter
    def query_online(self, query_online: str | None) -> None:
        self._query_online = query_online

    @prefix.setter
    def prefix(self, prefix: str | None) -> None:
        self._prefix = prefix
